################################################################################
# Copyright (c) 2014-2023, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Utilities for interfacing with accelerator hardware.

Both OpenCL and CUDA are supported, as well as multiple devices.

The modules :mod:`.cuda` and :mod:`.opencl` provide the abstraction layer,
but most code will not import these directly (and it might not be possible
to import them). Instead, use :func:`create_some_context` to set up a context
on whatever device is available.

Attributes
----------
have_cuda : bool
    True if PyCUDA could be imported (does not guarantee any CUDA devices)
have_opencl : bool
    True if PyOpenCL could be imported (does not guarantee any OpenCL devices)
"""

import re
import os
import sys
import io
from abc import ABC, abstractmethod
import itertools
from collections import OrderedDict
from typing import (List, Tuple, Dict, Set, Mapping, MutableMapping, Sequence,
                    Callable, Iterable, Optional, Union, TypeVar, Generic, Any,
                    cast, overload, TYPE_CHECKING)

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any     # type: ignore
import mako.lexer
from mako.template import Template
from mako.lookup import TemplateLookup
if sys.version_info >= (3, 9):
    import importlib.resources as importlib_resources
else:
    import importlib_resources

from .abc import (AbstractContext, AbstractDevice, AbstractCommandQueue,  # noqa: F401
                  AbstractTuningCommandQueue, AbstractProgram)

try:
    import pycuda.driver
    from . import cuda
    have_cuda = True
except ImportError:
    have_cuda = False
try:
    from . import opencl
    have_opencl = True
except ImportError:
    have_opencl = False

if TYPE_CHECKING:
    import graphviz


_Slice = Union[int, slice, None, Tuple[Union[int, slice, None], ...]]
_B = TypeVar('_B')     # buffer type
_RB = TypeVar('_RB')   # raw buffer type
_RS = TypeVar('_RS')   # raw buffer for SVM
_D = TypeVar('_D', bound='AbstractDevice')
_P = TypeVar('_P', bound='AbstractProgram')
_Q = TypeVar('_Q', bound='AbstractCommandQueue')
_TQ = TypeVar('_TQ', bound='AbstractTuningCommandQueue')


def divup(x: int, y: int) -> int:
    """Divide x by y and round the result upwards."""
    return (x + y - 1) // y


def roundup(x: int, y: int) -> int:
    """Round x up to the next multiple of y."""
    return divup(x, y) * y


class LinenoLexer(mako.lexer.Lexer):
    """Insert #line directives into the source code.

    It is used by passing `lexer_cls` to the mako template constructor.
    """

    def __init__(self, *args, **kw) -> None:
        super().__init__(*args, **kw)
        self.preprocessor.insert(0, self._lineno_preproc)

    @classmethod
    def _escape_filename(cls, filename: str) -> str:
        """Escape a string for the C preprocessor."""
        return '"' + re.sub(r'([\\"])', r'\\\1', filename) + '"'

    def _lineno_preproc(self, source: str) -> str:
        """Insert #line directives between every line in `source` and return the result."""
        if self.filename is not None:
            escaped_filename = self._escape_filename(self.filename)
        else:
            escaped_filename = self._escape_filename('<string>')
        lines = source.split('\n')
        # If the last line is \n-terminated, it will cause an empty
        # final string in lines
        if len(lines) > 0 and lines[-1] == '':
            lines.pop()
        out = []
        for i, line in enumerate(lines):
            out.append('#line {} {}\n'.format(i + 1, escaped_filename))
            out.append(line + '\n')
        return ''.join(out)


def _make_lookup(extra_dirs: List[Union[str, os.PathLike]]) -> TemplateLookup:
    dirs = extra_dirs + [importlib_resources.files("katsdpsigproc")]
    return TemplateLookup(dirs, lexer_cls=LinenoLexer, strict_undefined=True)


# Cached for convenience
_lookup = _make_lookup([])


def build(context: AbstractContext, name: str,
          render_kws: Optional[Mapping[str, Any]] = None,
          extra_dirs: Optional[List[Union[str, os.PathLike]]] = None,
          extra_flags: Optional[List[str]] = None,
          source: Optional[str] = None) -> AbstractProgram:
    """Build a source module from a mako template.

    Parameters
    ----------
    context
        Context for which to compile the code
    name
        Source file name, relative to the katsdpsigproc module
    render_kws
        Keyword arguments to pass to mako
    extra_dirs
        Extra directories to search for source code
    extra_flags
        Flags to pass to the compiler
    source
        If specified, provides the source, overriding `name`

    Returns
    -------
    program
        Compiled module
    """
    if render_kws is None:
        render_kws = {}
    if extra_flags is None:
        extra_flags = []
    if extra_dirs is None:
        lookup = _lookup
    else:
        lookup = _make_lookup(extra_dirs)
    if source is not None:
        template = Template(source, lookup=lookup, lexer_cls=LinenoLexer)
    else:
        template = lookup.get_template(name)
    rendered_source = template.render(
        simd_group_size=context.device.simd_group_size,
        **render_kws)
    return context.compile(rendered_source, extra_flags)


def all_devices() -> List[AbstractDevice]:
    """Return a list of all discovered devices."""
    devices = []     # type: List[AbstractDevice]
    if have_cuda:
        pycuda.driver.init()
        devices.extend(cuda.Device.get_devices())
    if have_opencl:
        devices.extend(opencl.Device.get_devices())
    return devices


def candidate_devices(device_filter: Optional[Callable[[AbstractDevice], bool]] = None) \
        -> Sequence[AbstractDevice]:
    """Get devices that are considered for :func:`create_some_context`.

    Refer to :func:`create_some_context` for documentation of how this list is
    affected by `device_filter` and environment variables. If no matching
    devices are found, returns an empty list. If an environment variable is
    out of range, raises :exc:`RuntimeError`.
    """

    def parse_id(envar: str) -> Optional[int]:
        if envar in os.environ:
            try:
                num = int(os.environ[envar])
                if num >= 0:
                    return num
            except ValueError:
                pass
        return None

    def parse_id_list(envar: str, max_parts: int) -> Optional[List[int]]:
        """Split an environment variable of the form a[:b[:c...]] into its components.

        If the variable is not validly formatted, returns `None`.
        """
        if envar in os.environ:
            try:
                fields = os.environ[envar].split(':', max_parts)
                out = []
                for value in fields:
                    num = int(value)
                    if num < 0:
                        raise ValueError('')
                    out.append(num)
                if not out:
                    raise ValueError('')
                return out
            except ValueError:
                pass
        return None

    cuda_id = None         # type: Optional[int]
    opencl_id = None       # type: Optional[List[int]]
    device_id = None       # type: Optional[int]

    device_id = parse_id('KATSDPSIGPROC_DEVICE')
    if device_id is None:
        cuda_id = parse_id('CUDA_DEVICE')
        if cuda_id is None:
            # Fields are platform number and device number
            opencl_id = parse_id_list('PYOPENCL_CTX', 2)

    cuda_devices = []      # type: Sequence[cuda.Device]
    opencl_devices = []    # type: Sequence[Sequence[opencl.Device]]
    devices = []           # type: List[AbstractDevice]
    if have_cuda:
        pycuda.driver.init()
        cuda_devices = cuda.Device.get_devices()
    if have_opencl:
        opencl_devices = opencl.Device.get_devices_by_platform()
    if device_filter is not None:
        cuda_devices = list(filter(device_filter, cuda_devices))
        opencl_devices = [list(filter(device_filter, platform)) for platform in opencl_devices]

    try:
        if cuda_id is not None:
            devices.append(cuda_devices[cuda_id])
        elif opencl_id is not None:
            if len(opencl_id) == 1:
                devices.extend(opencl_devices[opencl_id[0]])
            else:
                devices.append(opencl_devices[opencl_id[0]][opencl_id[1]])
        else:
            devices = list(itertools.chain(cuda_devices, *opencl_devices))
            if device_id is not None:
                devices = [devices[device_id]]
    except IndexError:
        raise RuntimeError('Out-of-range device selected')
    return devices


def create_some_context(
        interactive: bool = True,
        device_filter: Optional[Callable[[AbstractDevice], bool]] = None) -> AbstractContext:
    """Create a single-device context, selecting a device automatically.

    This is similar to `pyopencl.create_some_context`. A number of environment
    variables can be set to limit the choice to a single device:

     - KATSDPSIGPROC_DEVICE: device number from amongst all devices
     - CUDA_DEVICE: CUDA device number (compatible with PyCUDA)
     - PYOPENCL_CTX: OpenCL platform and optionally device number (compatible with PyOpenCL)

    The first of these that is encountered takes effect. If it does not exist,
    an exception is thrown.

    Parameters
    ----------
    interactive
        If true, and `sys.stdin.isatty()` is true, and there are multiple
        choices, it will prompt the user. Otherwise, it will choose the first
        available device, favouring CUDA over OpenCL, then GPU over
        accelerators over other OpenCL devices.
    device_filter
        If specified, each device in turn is passed to it, and it must return
        True to keep the device as a candidate or False to reject it.

    Raises
    ------
    RuntimeError
        If no device could be found or the user made an invalid selection
    """
    def key(device: AbstractDevice) -> int:
        if device.is_cuda:
            return 100
        elif device.is_gpu:
            return 50
        elif device.is_accelerator:
            return 40
        else:
            return 30

    devices = candidate_devices(device_filter)
    if not devices:
        raise RuntimeError('No compute devices found')

    if interactive and len(devices) > 1 and sys.stdin.isatty():
        print("Select device:")
        for i, device in enumerate(devices):
            print(f"    [{i}]: {device.name} ({device.platform_name})")
        print()
        choice_str = input('Enter selection: ')
        try:
            choice = int(choice_str)
            if choice < 0:
                raise IndexError   # Otherwise Python's negative indexing kicks in
            device = devices[choice]
        except (ValueError, IndexError):
            raise RuntimeError('Invalid device number')
    else:
        device = max(devices, key=key)

    return device.make_context()


class HostArray(np.ndarray):
    """A restricted array class that can be used to initialise a :class:`DeviceArray`.

    It uses C ordering and allows padding, which is always in units of the
    dtype. It optionally uses pinned memory to allow fast transfer to and from
    device memory.

    Because of limitations in numpy and PyCUDA (which do not support
    non-contiguous memory very well), it works by taking a slice from
    the origin of contiguous storage, and using that contiguous storage
    in host-device copies. While one can create views, those views
    cannot be used in fast copies because there is no way to know
    whether the view is anchored at the origin.

    See the numpy documentation on subclassing for an explanation of
    why it is written the way it is.

    Parameters
    ----------
    shape
        Shape for the array
    dtype
        Data type for the array
    padded_shape
        Total size of memory allocation (defaults to `shape`)
    context
        If specified, the memory will be allocated in a way that allows
        efficient copies to and from this context.
    """

    def __new__(cls, shape: Tuple[int, ...], dtype: DTypeLike,
                padded_shape: Optional[Tuple[int, ...]] = None,
                context: Optional[AbstractContext] = None):
        if padded_shape is None:
            padded_shape = shape
        assert len(padded_shape) == len(shape)
        assert np.all(np.greater_equal(padded_shape, shape))
        if context is not None:
            owner = context.allocate_pinned(padded_shape, dtype)
        else:
            owner = np.empty(padded_shape, dtype)
        if shape:
            index = tuple([slice(0, x) for x in shape])
            obj = owner[index].view(HostArray)
        else:
            obj = owner.view(HostArray)
        obj._owner = owner
        obj.padded_shape = padded_shape
        return obj

    @classmethod
    def safe(cls, obj: np.ndarray) -> bool:
        """Determine whether self can be copied to/from a device directly."""
        try:
            return obj._owner is not None  # type: ignore
        except AttributeError:
            return False

    @overload
    @classmethod
    def padded_view(cls, obj: 'HostArray') -> np.ndarray:
        ...

    @overload
    @classmethod
    def padded_view(cls, obj: np.ndarray) -> Optional[np.ndarray]:    # noqa: F811
        ...

    @classmethod
    def padded_view(cls, obj: np.ndarray) -> Optional[np.ndarray]:    # noqa: F811
        """Retrieve the view of the full memory without padding.

        Returns `None` if `cls.safe(obj)` is `False`.
        """
        try:
            return obj._owner  # type: ignore
        except AttributeError:
            return None

    def __array_finalize__(self, obj: 'HostArray') -> None:
        if obj is not None:
            # View casting or created from a template: we cannot vouch for it.
            self._owner = None
            if isinstance(obj, HostArray):
                self.padded_shape = obj.padded_shape    # type: Tuple[int, ...]
            else:
                self.padded_shape = None


class DeviceArray:
    """A light-weight array-like wrapper around a device buffer.

    It that handles padding better than PyCUDA (which has very poor support).

    It only supports C-order arrays where the inner-most dimension is
    contiguous. Transfers are designed to use an :class:`HostArray` of the
    same shape and padding, but fall back to using a copy when
    necessary.

    Parameters
    ----------
    context
        Context in which to allocate the memory
    shape
        Shape for the usable data
    dtype
        Data type
    padded_shape
        Shape for memory allocation (defaults to `shape`)
    raw
        If specified, provides the backing memory. It must be created from
        the `allocate_raw` method of an allocator.
    """

    def __init__(self, context: AbstractContext, shape: Tuple[int, ...],
                 dtype: DTypeLike, padded_shape: Optional[Tuple[int, ...]] = None,
                 raw: Any = None) -> None:
        if padded_shape is None:
            padded_shape = shape
        assert len(shape) == len(padded_shape)
        assert np.all(np.greater_equal(padded_shape, shape))
        self._shape = shape
        self._dtype: np.dtype = np.dtype(dtype)
        self.padded_shape = padded_shape
        self.context = context
        self.buffer = context.allocate(padded_shape, dtype, raw)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape

    @property
    def dtype(self) -> np.dtype:
        return self._dtype

    @property
    def ndim(self) -> int:
        """Return number of dimensions."""
        return len(self.shape)

    @property
    def strides(self) -> Tuple[int, ...]:
        """Return strides, as in numpy."""
        ans = [self.dtype.itemsize]
        for i in range(len(self.padded_shape) - 1, 0, -1):
            ans.append(ans[-1] * self.padded_shape[i])
        return tuple(reversed(ans))

    def _copyable(self, ary: np.ndarray) -> bool:
        """Whether `ary` can be copied to/from this array directly."""
        return (HostArray.safe(ary)
                and ary.dtype == self.dtype
                and ary.shape == self.shape
                and ary.padded_shape == self.padded_shape)  # type: ignore

    @classmethod
    def _contiguous(cls, ary: HostArray) -> np.ndarray:
        """Return a contiguous view of a copyable array.

        The return value is suitable for passing to PyCUDA or PyOpenCL
        functions (which require a contiguous view).
        """
        return HostArray.padded_view(ary)

    def empty_like(self) -> HostArray:
        """Return an array-like object that can be efficiently copied."""
        return HostArray(self.shape, self.dtype, self.padded_shape, context=self.context)

    def asarray_like(self, ary: np.ndarray) -> HostArray:
        """Return an array with the same content as `ary`, but the same memory layout as `self`."""
        assert ary.shape == self.shape
        if self._copyable(ary):
            return ary  # type: ignore
        tmp = self.empty_like()
        np.copyto(tmp, ary, casting='no')
        return tmp

    def set(self, command_queue: AbstractCommandQueue, ary: np.ndarray) -> None:
        """Copy synchronously from `ary` to self."""
        ary = self.asarray_like(ary)
        command_queue.enqueue_write_buffer(self.buffer, self._contiguous(ary))

    def get(self, command_queue: AbstractCommandQueue,
            ary: Optional[np.ndarray] = None) -> np.ndarray:
        """Copy synchronously from self to `ary`.

        If `ary` is None, or if it is not suitable as a target, the copy is to
        a newly-allocated :class:`HostArray`. The actual target is returned.
        """
        if ary is None or not self._copyable(ary):
            ary = self.empty_like()
        assert isinstance(ary, HostArray)
        command_queue.enqueue_read_buffer(self.buffer, self._contiguous(ary))
        return ary

    def set_async(self, command_queue: AbstractCommandQueue, ary: np.ndarray) -> None:
        """Copy asynchronously from `ary` to self."""
        ary = self.asarray_like(ary)
        command_queue.enqueue_write_buffer(
            self.buffer, self._contiguous(ary), blocking=False)

    def get_async(self, command_queue: AbstractCommandQueue,
                  ary: Optional[np.ndarray] = None) -> np.ndarray:
        """Copy asynchronously from self to `ary` (see :meth:`get`)."""
        if ary is None or not self._copyable(ary):
            ary = self.empty_like()
        assert isinstance(ary, HostArray)
        command_queue.enqueue_read_buffer(
            self.buffer, self._contiguous(ary), blocking=False)
        return ary

    @classmethod
    def _canonical_slice(cls, region: _Slice, shape: Tuple[int, ...],
                         strides: Tuple[int, ...]) -> Tuple[int, Tuple[int, ...], Tuple[int, ...]]:
        """Transform a slice selection into a form that is easier to consume internally.

        See :meth:`copy_region` for a description of what slice
        expressions are supported.

        Parameters
        ----------
        region : int, slice, `np.newaxis` or tuple of same
            Index expression generated by `np.s_` or `np.index_exp`
        shape : sequence
            Shape of the array being indexed
        strides : sequence
            Strides of the array being indexed

        Returns
        -------
        origin : int
            Byte offset for the first selected element
        shape : tuple of int
            Shape of the region to copy
        strides : tuple of int
            Byte stride between elements along each axis of `shape`. Each stride
            is a multiple of the following ones, except that a dimension of
            length 1 may have a stride of 0.
        """
        origin = 0
        copy_shape = []      # type: List[int]
        copy_strides = []
        axis = 0
        if not isinstance(region, tuple):
            region = (region,)
        for index in region:
            if index is np.newaxis:
                copy_shape.append(1)
                copy_strides.append(0)
            elif isinstance(index, slice):
                if axis >= len(shape):
                    raise IndexError('Too many axes in index expression')
                start, stop, stride = index.indices(shape[axis])
                if stride <= 0:
                    raise IndexError('Only positive strides are supported')
                origin += start * strides[axis]
                copy_shape.append((stop - start) // stride)
                if copy_shape[-1] <= 0:
                    raise IndexError('Empty slice selection')
                copy_strides.append(stride * strides[axis])
                axis += 1
            elif isinstance(index, int):
                if axis >= len(shape):
                    raise IndexError('Too many axes in index expression')
                if index < 0:
                    index += shape[axis]
                if index < 0 or index >= shape[axis]:
                    raise IndexError('Index out of range')
                origin += index * strides[axis]
                axis += 1
            else:
                raise TypeError('Invalid type in slice: {}'.format(type(index)))
        while axis < len(shape):
            copy_shape.append(shape[axis])
            copy_strides.append(strides[axis])
            axis += 1
        return origin, tuple(copy_shape), tuple(copy_strides)

    @classmethod
    def _region_transfer_params(
            cls, src: Any, dest: Any,
            src_region: _Slice,
            dest_region: _Slice) -> Tuple[int, int, Tuple[int, ...],
                                          Tuple[int, ...], Tuple[int, ...]]:
        """Compute offsets, strides and shape for a copy.

        Where possible, the dimension of the copy is reduced by exploiting
        contiguity. When interpreting the return values, the arrays should be
        thought of as byte arrays, ignoring any shape, padding or dtype of the
        originals.

        Parameters
        ----------
        src,dest : array-like
            Source and destination buffers, with `dtype, `shape` and `strides`
            attributes
        src_region,dest_region : int, slice, `np.newaxis` or tuple of same
            Index expression generated by `np.s_` or `np.index_exp`

        Returns
        -------
        src_origin,dest_origin : int
            Byte offset into the source/destination array for the start of the copy
        shape : tuple
            Shape for the copy, which may have fewer dimensions than the arrays.
        src_strides,dest_strides : tuple
            Byte strides for source/destination.

        Raises
        ------
        TypeError
            if the source and destination do not have the same dtype
        ValueError
            if the source and destination regions are not the same shape
        IndexError
            if the source or destination regions are unsupported or out-of-range
        """
        if src.dtype != dest.dtype:
            raise TypeError(f'dtypes do not match ({src.dtype} and {dest.dtype})')
        src_origin, src_shape, src_strides = cls._canonical_slice(
            src_region, src.shape, src.strides)
        dest_origin, dest_shape, dest_strides = cls._canonical_slice(
            dest_region, dest.shape, dest.strides)
        if src_shape != dest_shape:
            raise ValueError('Source and destination shapes for the copy do not match')
        # Search for axes that can be collapsed together
        new_shape = [src.dtype.itemsize]
        new_src_strides = [1]
        new_dest_strides = [1]
        for axis in range(len(src_shape) - 1, -1, -1):
            if src_shape[axis] == 1:
                continue    # Can just ignore this axis
            if (src_strides[axis] == new_shape[-1] * new_src_strides[-1]
                    and dest_strides[axis] == new_shape[-1] * new_dest_strides[-1]):
                new_shape[-1] *= src_shape[axis]
            else:
                new_shape.append(src_shape[axis])
                new_src_strides.append(src_strides[axis])
                new_dest_strides.append(dest_strides[axis])
        return (src_origin, dest_origin,
                tuple(new_shape), tuple(new_src_strides), tuple(new_dest_strides))

    @classmethod
    def _transfer_region(cls, func: Callable,
                         buffer1: Any, buffer2: Any,
                         origin1: int, origin2: int,
                         shape: Tuple[int, ...],
                         strides1: Tuple[int, ...], strides2: Tuple[int, ...],
                         **kwargs) -> None:
        """Wrap the command-queue's copy and handle high dimensions by looping."""
        if len(shape) > 3:
            for i in range(shape[-1]):
                cls._transfer_region(
                    func, buffer1, buffer2,
                    origin1 + strides1[-1] * i,
                    origin2 + strides2[-1] * i,
                    shape[:-1], strides1[:-1], strides2[:-1], **kwargs)
        else:
            func(buffer1, buffer2, origin1, origin2, shape, strides1, strides2, **kwargs)

    def copy_region(self, command_queue: AbstractCommandQueue, dest: 'DeviceArray',
                    src_region: _Slice, dest_region: _Slice) -> None:
        """Perform a device-to-device copy of a subregion of `self` to `dest`.

        If the source and destination memory overlap, the result is undefined.

        The regions to copy are specified using a subset of numpy array
        indexing syntax. The following are supported:

        - slices with positive strides
        - integers
        - :data:`np.newaxis <numpy.newaxis>`
        - If fewer indices than axes are specified, all elements on the
          remaining axes are used.

        Ellipses are not yet supported, but it would be straightforward to add
        support.

        Parameters
        ----------
        command_queue
            Command queue for the asynchronous operation.
        dest
            Target of the copy
        src_region,dest_region
            Index expressions constructed by :data:`np.s_ <numpy.s_>` or
            :data:`!np.index_exp`.

        Raises
        ------
        TypeError
            if the source and destination do not have the same dtype
        ValueError
            if the source and destination regions select different shapes
        IndexError
            if the source or destination regions are unsupported or out-of-range
        """
        src_origin, dest_origin, shape, src_strides, dest_strides = \
            self._region_transfer_params(self, dest, src_region, dest_region)
        self._transfer_region(
            command_queue.enqueue_copy_buffer_rect,
            self.buffer, dest.buffer, src_origin, dest_origin,
            shape, src_strides, dest_strides)

    def get_region(self, command_queue: AbstractCommandQueue, ary: np.ndarray,
                   device_region: _Slice, ary_region: _Slice, blocking: bool = True) -> None:
        """Perform a device-to-host copy of a subregion of `self` to `ary`.

        See :meth:`~DeviceArray.copy_region` for a description of how regions
        are specified.

        Parameters
        ----------
        command_queue
            Command queue for the operation.
        ary
            Target of the copy
        device_region,ary_region
            Index expressions constructed by :data:`np.s_ <numpy.s_>` or
            :data:`!np.index_exp`, to specify the source and target regions.
        blocking
            If false, the operation will be asynchronous.

        Raises
        ------
        TypeError
            if the source and destination do not have the same dtype
        ValueError
            if the source and destination regions select different shapes
        ValueError
            if `ary` is not a :class:`HostArray` or is a view of one
        IndexError
            if the source or destination regions are unsupported or out-of-range
        """
        if not HostArray.safe(ary):
            raise ValueError('Target region is not suitable for device-to-host copy')
        assert isinstance(ary, HostArray)
        device_origin, ary_origin, shape, device_strides, ary_strides = \
            self._region_transfer_params(self, ary, device_region, ary_region)
        self._transfer_region(
            command_queue.enqueue_read_buffer_rect,
            self.buffer, self._contiguous(ary), device_origin, ary_origin,
            shape, device_strides, ary_strides, blocking=blocking)

    def set_region(self, command_queue: AbstractCommandQueue, ary: np.ndarray,
                   device_region: _Slice, ary_region: _Slice, blocking: bool = True) -> None:
        """Perform a host-to-device copy of a subregion `ary` to `self`.

        See :meth:`~DeviceArray.copy_region` for a description of how regions
        are specified.

        Parameters
        ----------
        command_queue
            Command queue for the operation.
        ary
            Source of the copy
        device_region,ary_region
            Index expressions constructed by :data:`np.s_ <numpy.s_>` or
            :data:`!np.index_exp`, to specify the source and target regions.
        blocking
            If false, the operation will be asynchronous.

        Raises
        ------
        TypeError
            if the source and destination do not have the same dtype
        ValueError
            if the source and destination regions select different shapes
        IndexError
            if the source or destination regions are unsupported or out-of-range
        """
        if not HostArray.safe(ary):
            tmp = HostArray(ary[ary_region].shape, ary.dtype, context=self.context)
            np.copyto(tmp, ary[ary_region], casting='no')
            ary = tmp
            ary_region = np.s_[()]
        assert isinstance(ary, HostArray)
        ary_origin, device_origin, shape, ary_strides, device_strides = \
            self._region_transfer_params(ary, self, ary_region, device_region)
        self._transfer_region(
            command_queue.enqueue_write_buffer_rect,
            self.buffer, self._contiguous(ary), device_origin, ary_origin,
            shape, device_strides, ary_strides, blocking=blocking)

    def zero(self, command_queue: AbstractCommandQueue) -> None:
        """Memset with zeros (asynchronously)."""
        command_queue.enqueue_zero_buffer(self.buffer)


class SVMArray(HostArray, DeviceArray):
    """An array that uses shared virtual memory to be accessible from both the host and the device.

    Shared virtual memory is also known as managed memory (in CUDA).

    It should not be used as a source or target for copies to the device, since
    it already resides on the device.

    Due to limitations in PyOpenCL, this is currently only available for CUDA,
    and CUDA's restrictions on managed memory apply. Only the base array (not
    views) can be passed to kernels.

    Parameters
    ----------
    context
        Context in which to allocate the memory
    shape
        Shape for the array
    dtype
        Data type for the array
    padded_shape
        Total size of memory allocation (defaults to `shape`)
    raw : PyCUDA DeviceAllocation or PyOpenCL Buffer, optional
        If specified, provides the backing memory
    """

    def __new__(cls, context: AbstractContext, shape: Tuple[int, ...], dtype: DTypeLike,
                padded_shape: Optional[Tuple[int, ...]] = None, raw: Any = None) -> 'SVMArray':
        if padded_shape is None:
            padded_shape = shape
        assert len(padded_shape) == len(shape)
        assert np.all(np.greater_equal(padded_shape, shape))
        owner = context.allocate_svm(padded_shape, dtype, raw=raw)
        if shape:
            index = tuple([slice(0, x) for x in shape])
            obj = owner[index].view(SVMArray)
        else:
            obj = owner.view(SVMArray)
        obj._owner = owner
        obj.padded_shape = padded_shape
        obj.context = context
        return obj

    def __init__(self, *args, **kwargs):
        # Suppress DeviceArray's constructor, since we did everything in __new__
        pass

    @property
    def buffer(self) -> np.ndarray:
        return self._owner  # type: ignore

    def _copyable(self, ary: np.ndarray) -> bool:
        """Whether `ary` can be copied to/from this array directly."""
        return (ary.dtype == self.dtype
                and ary.shape == self.shape)

    def set(self, command_queue: AbstractCommandQueue, ary: np.ndarray) -> None:
        """Copy synchronously from `ary` to self.

        For SVMArray, this is a CPU copy.
        """
        # Ensure that it synchronises with previous work in the queue
        command_queue.finish()
        self[...] = ary

    def get(self, command_queue: AbstractCommandQueue,
            ary: Optional[np.ndarray] = None) -> np.ndarray:
        """Copy synchronously copy from self to `ary`.

        If `ary` is None, or if it is not suitable as a target, the copy is to
        a newly allocated :class:`HostArray`. The actual target is returned.
        For SVMArray, this is a CPU copy.
        """
        # Ensure that it synchronises with previous work in the queue
        command_queue.finish()
        if ary is None or not self._copyable(ary):
            return self.copy()
        else:
            ary[:] = self
            return ary

    def set_async(self, command_queue: AbstractCommandQueue, ary: np.ndarray) -> None:
        """Copy asynchronously from `ary` to self.

        This is implemented synchronously for SVMArray, but exists for
        compatibility.
        """
        self.set(command_queue, ary)

    def get_async(self, command_queue: AbstractCommandQueue,
                  ary: Optional[np.ndarray] = None) -> np.ndarray:
        """Copy asynchronously from self to `ary` (see `get`).

        This is implemented synchronously for SVMArray, but exists for
        compatibility.
        """
        return self.get(command_queue, ary)

    def set_region(self, command_queue: AbstractCommandQueue, ary: np.ndarray,
                   device_region: _Slice, ary_region: _Slice, blocking: bool = True) -> None:
        # Ensure that it synchronises with previous work in the queue
        command_queue.finish()
        self[device_region] = ary[ary_region]

    def get_region(self, command_queue: AbstractCommandQueue, ary: np.ndarray,
                   device_region: _Slice, ary_region: _Slice, blocking: bool = True) -> None:
        # Ensure that it synchronises with previous work in the queue
        command_queue.finish()
        ary[ary_region] = self[device_region]


class AbstractAllocator(ABC, Generic[_RB]):
    """Interface for allocating device memory."""

    context = None    # type: AbstractContext

    @abstractmethod
    def allocate(self, shape: Tuple[int, ...], dtype: DTypeLike,
                 padded_shape: Optional[Tuple[int, ...]] = None,
                 raw: Optional[_RB] = None) -> DeviceArray:
        pass

    @abstractmethod
    def allocate_raw(self, n_bytes: int) -> _RB:
        pass


class DeviceAllocator(Generic[_B, _RB, _RS, _D, _P, _Q, _TQ], AbstractAllocator[_RB]):
    """Allocate :class:`DeviceArray` objects from a context."""

    def __init__(self, context: AbstractContext[_B, _RB, _RS, _D, _P, _Q, _TQ]) -> None:
        self.context = context

    def allocate(self, shape: Tuple[int, ...], dtype: DTypeLike,
                 padded_shape: Optional[Tuple[int, ...]] = None,
                 raw: Optional[_RB] = None) -> DeviceArray:
        return DeviceArray(self.context, shape, dtype, padded_shape, raw)

    def allocate_raw(self, n_bytes: int) -> _RB:
        return self.context.allocate_raw(n_bytes)


class SVMAllocator(Generic[_B, _RB, _RS, _D, _P, _Q, _TQ], AbstractAllocator[_RS]):
    """Allocate :class:`SVMArray` objects from a context."""

    def __init__(self, context: AbstractContext[_B, _RB, _RS, _D, _P, _Q, _TQ]) -> None:
        self.context = context

    def allocate(self, shape: Tuple[int, ...], dtype: DTypeLike,
                 padded_shape: Optional[Tuple[int, ...]] = None,
                 raw: Optional[_RS] = None) -> SVMArray:
        return SVMArray(self.context, shape, dtype, padded_shape, raw)

    def allocate_raw(self, n_bytes: int) -> _RS:
        return self.context.allocate_svm_raw(n_bytes)


class Dimension:
    """A single dimension of an :class:`IOSlot`.

    It represents padding and alignment requirements. Instances can be linked
    together, with all linked instances (whether directly or indirectly linked)
    exposing the intersection of their requirements. Internally this is
    represented using a union-find tree with path compression.

    Note that `min_padded_round` and `alignment` provide similar but
    different requirements. The former is used to determine a minimum amount
    of padding, but any larger amount is acceptable, even if not a multiple of
    `min_padded_round`. The latter is stricter, always requiring a multiple of
    this value. The latter is not expected to be frequently used except when
    type-punning.

    Dimensions can also be frozen, which is done by :class:`IOSlot` once a
    buffer is bound. This prevents the requirements from changing later (via
    linking or otherwise) and hence invalidating the bound buffer, or
    allowing two buffers with a shared dimension but different strides to
    be used.

    Parameters
    ----------
    size
        Size of the actual data, before padding
    min_padded_round
        The `min_padded_size` will be computed by rounding `size` up to a
        multiple of this.
    min_padded_size
        Minimum size of the padded allocation, overriding `min_padded_round`.
    alignment
        Padded size is required to be a multiple of this.
        This must be a power of 2.
    align_dtype
        If specified, it is a hint that this data type is the fastest-varying
        axis of a multidimensional array. The padded size may be chosen to be
        such that the stride is a multiple of `ALIGN_BYTES`, to ensure
        efficient access by GPUs. The hint will be ignored if `exact` is true.
    exact
        If true, padding is forbidden
    """

    ALIGN_BYTES = 128

    @classmethod
    def _is_power2(cls, value: int) -> bool:
        return value > 0 and (value & (value - 1)) == 0

    def __init__(self, size: int,
                 min_padded_round: Optional[int] = None, min_padded_size: Optional[int] = None,
                 alignment: int = 1, align_dtype: Optional[DTypeLike] = None,
                 exact: bool = False) -> None:
        if min_padded_size is None:
            if min_padded_round is not None:
                min_padded_size = roundup(size, min_padded_round)
            else:
                min_padded_size = size

        if not self._is_power2(alignment):
            raise ValueError('alignment is not a power of 2')
        if min_padded_size < size:
            raise ValueError('padded size is less than size')

        self._parent = None     # type: Optional[Dimension]
        self._size = size
        self._min_padded_size = min_padded_size
        self._alignment = alignment
        self._alignment_hint = alignment
        self._exact = exact
        self._frozen = False
        if align_dtype is not None:
            self.add_align_dtype(align_dtype)

    def _root(self) -> 'Dimension':
        if self._parent is None:
            return self
        else:
            self._parent = self._parent._root()
            return self._parent

    @property
    def size(self) -> int:
        return self._root()._size

    @property
    def min_padded_size(self) -> int:
        return self._root()._min_padded_size

    @property
    def alignment(self) -> int:
        return self._root()._alignment

    @property
    def alignment_hint(self) -> int:
        return self._root()._alignment_hint

    @property
    def exact(self) -> bool:
        return self._root()._exact

    @property
    def frozen(self) -> bool:
        return self._root()._frozen

    def required_padded_size(self) -> int:
        """Return padded size required to satisfy this dimension."""
        root = self._root()
        size = roundup(root._min_padded_size, root._alignment)
        # If the size is less than the alignment hint, it could
        # waste a huge amount of memory to pad it
        if not root._exact and size >= root._alignment_hint:
            size = roundup(size, root._alignment_hint)
        return size

    def valid(self, padded_size: int) -> bool:
        """Return whether `size` is a valid padded size."""
        root = self._root()
        if root._exact:
            return padded_size == root.required_padded_size()
        else:
            return (padded_size >= root._min_padded_size
                    and padded_size % root._alignment == 0)

    def add_align_dtype(self, dtype: DTypeLike) -> None:
        """Add an alignment hint.

        Indicates that this will be used with an array whose fastest-varying
        dimension is of type `dtype`. If the size is not a power of 2, it is
        ignored.
        """
        if self.frozen:
            raise ValueError('cannot modify a frozen requirement')
        itemsize = np.dtype(dtype).itemsize
        if self._is_power2(itemsize):
            root = self._root()
            root._alignment_hint = max(root._alignment_hint, self.ALIGN_BYTES // itemsize)

    def link(self, other: 'Dimension') -> None:
        """Make both `self` and `other` reference a single shared requirement.

        Raises
        ------
        ValueError
            If the resulting requirement is exact and unsatisfiable
        """
        root1 = self._root()
        root2 = other._root()
        if root1._frozen or root2._frozen:
            raise ValueError('cannot link frozen requirements')
        if root1 is root2:
            return
        if root1._size != root2._size:
            raise ValueError('sizes are incompatible')
        if root1._exact:
            actual1 = root1.required_padded_size()
            if not root2.valid(actual1):
                raise ValueError('linked requirement is unsatisfiable')
        if root2._exact:
            actual2 = root2.required_padded_size()
            if not root2.valid(actual2):
                raise ValueError('linked requirement is unsatisfiable')
        root1._min_padded_size = max(root1._min_padded_size, root2._min_padded_size)
        root1._alignment = max(root1._alignment, root2._alignment)
        root1._alignment_hint = max(root1._alignment_hint, root2._alignment_hint)
        root1._exact = root1._exact or root2._exact
        root2._parent = root1
        # Clear properties on the child, so that they can't accidentally be used
        # _size can stay as it is immutable
        del root2._min_padded_size
        del root2._alignment
        del root2._alignment_hint
        del root2._exact

    def freeze(self) -> None:
        """Prevent further modifications."""
        self._root()._frozen = True


class IOSlotBase(ABC):
    """An input/output slot of an operation.

    A slot can be bound to storage, or can allocate storage itself. This base
    class is untyped and unshaped, so in most cases one will use
    :class:`IOSlot` instead.

    Slots are arranged in a tree, and only the root can be manipulated
    directly. The entire tree shares the same storage.
    A slot cannot be reattached to a new parent.
    """

    def __init__(self) -> None:
        self.is_root = True

    def check_root(self) -> None:
        """Check whether this slot is a root slot, and raise an exception if not."""
        if not self.is_root:
            raise ValueError('not a root slot')

    @abstractmethod
    def required_bytes(self) -> int:
        """Return number of bytes of device storage required."""

    @abstractmethod
    def is_bound(self):
        """Return whether storage is currently attached to this slot."""

    def attachable(self) -> bool:
        """Return whether this slot can be attached as a child to another."""
        return self.is_root and not self.is_bound()

    @abstractmethod
    def _allocate(self, allocator: AbstractAllocator[_RB], raw: Optional[_RB] = None,
                  *, bind: bool) -> Any:
        """Variant of :meth:`allocate` that does not check whether this is a root slot."""

    def allocate(self, allocator: AbstractAllocator[_RB], raw: Optional[_RB] = None,
                 *, bind: bool = True) -> Any:
        """Allocate and optionally bind a buffer satisfying the requirements.

        .. warning:: When `raw` is provided, there is no check that the storage is large enough.

        Parameters
        ----------
        allocator
            Memory allocator from which to obtain the memory
        raw : PyCUDA DeviceAllocation or PyOpenCL Buffer, optional
            Backing storage for the allocation
        bind
            If true (the default), the allocated buffer is immediately bound.
            If false, the buffer is returned and not bound, and can be bound
            later.

        Raises
        ------
        ValueError
            If this is not a root slot
        """
        self.check_root()
        return self._allocate(allocator, raw, bind=bind)

    @abstractmethod
    def _allocate_host(self, context: AbstractContext) -> HostArray:
        """Variant of :meth:`allocate_host` that does not check whether this is a root slot."""

    def allocate_host(self, context: AbstractContext) -> HostArray:
        """Allocate a HostArray compatible with this slot."""
        self.check_root()
        return self._allocate_host(context)


class IOSlot(IOSlotBase):
    """An input/output slot with type and shape information.

    It contains a reference to a buffer (initially `None`) and the shape, type,
    padding and alignment requirements for that buffer. It can allocate the
    buffer on request. Each dimension is represesented by a :class:`Dimension`.

    Slots are arranged in a tree, and only the root can be manipulated
    directly. A slot cannot be reattached to a new parent.

    Parameters
    ----------
    dimensions : tuple of int or :class:`Dimension`
        The dimensions of the slots. Integers are converted to
        :class:`Dimension` with no further requirements.
    dtype : numpy dtype
        Type of the data elements
    """

    @classmethod
    def _make_dimension(cls, dimension: Union[Dimension, int]) -> Dimension:
        if isinstance(dimension, Dimension):
            return dimension
        else:
            return Dimension(dimension)

    def __init__(self, dimensions: Tuple[Union[Dimension, int], ...], dtype: DTypeLike) -> None:
        super().__init__()
        self.dimensions = tuple([self._make_dimension(x) for x in dimensions])
        self.shape = tuple([x.size for x in self.dimensions])
        if len(self.dimensions) > 1:
            self.dimensions[-1].add_align_dtype(dtype)
        self.dtype: np.dtype = np.dtype(dtype)
        self.buffer = None       # type: Optional[DeviceArray]

    def is_bound(self) -> bool:
        return self.buffer is not None

    def validate(self, buffer: DeviceArray) -> None:
        """Check that `buffer` is suitable for binding.

        Parameters
        ----------
        buffer
            Buffer to validate (must not be `None`)

        Raises
        ------
        TypeError
            If the data type does not match
        ValueError
            If the dimensions or shape do not match
        ValueError
            If a padded size does not match :meth:`Dimension.required_padded_size`
        """
        if buffer.dtype != self.dtype:
            raise TypeError('dtype does not match')
        if len(buffer.shape) != len(self.shape):
            raise ValueError('number of dimensions does not match')
        for size, padded_size, dimension in zip(buffer.shape, buffer.padded_shape, self.dimensions):
            if size != dimension.size:
                raise ValueError('size does not match')
            if padded_size != dimension.required_padded_size():
                raise ValueError('padded size does not match')

    def _bind(self, buffer: Optional[DeviceArray]) -> None:
        """Variant of :meth:`bind` that does not check whether this is a root slot."""
        if buffer is not None:
            self.validate(buffer)
        self.buffer = buffer
        for dimension in self.dimensions:
            dimension.freeze()

    def bind(self, buffer: Optional[DeviceArray]) -> None:
        """Set the internal buffer reference.

        Always use this function rather than writing it directly.

        If the buffer is not `None`, it is validated (see :meth:`validate`).

        Parameters
        ----------
        buffer
            Buffer to store

        Raises
        ------
        ValueError
            If this is not a root slot
        """
        self.check_root()
        self._bind(buffer)

    def required_padded_shape(self) -> Tuple[int, ...]:
        """Return padded shape required to satisfy only this slot."""
        return tuple(x.required_padded_size() for x in self.dimensions)

    def required_bytes(self) -> int:
        return int(np.prod(self.required_padded_shape()) * self.dtype.itemsize)

    def _allocate(self, allocator: AbstractAllocator[_RB],
                  raw: Optional[_RB] = None, *, bind: bool = True) -> DeviceArray:
        buffer = allocator.allocate(self.shape, self.dtype, self.required_padded_shape(), raw=raw)
        if bind:
            self._bind(buffer)
        return buffer

    def _allocate_host(self, context: AbstractContext) -> HostArray:
        return HostArray(self.shape, self.dtype, self.required_padded_shape(), context=context)

    def allocate(self, allocator: AbstractAllocator[_RB],
                 raw: Optional[_RB] = None, *, bind: bool = True) -> DeviceArray:
        # Overloaded just to restrict the type signature
        return super().allocate(allocator, raw, bind=bind)


class CompoundIOSlot(IOSlot):
    """IO slot that owns multiple child slots, and presents the combined requirement.

    The children must all have the same type and shape. This is used for
    connecting a single buffer to multiple operations.

    Parameters
    ----------
    children
        Child slots

    Raises
    ------
    ValueError
        If `children` is empty, or if elements have inconsistent shapes
    ValueError
        If any child is not attachable
    TypeError
        If `children` contains inconsistent data types
    """

    def __init__(self, children: Iterable[IOSlot]) -> None:
        self.children = list(children)
        if not len(self.children):
            raise ValueError('empty child list')
        shape = self.children[0].shape
        dtype = self.children[0].dtype
        dimensions = self.children[0].dimensions
        # Validate consistency and compute combined requirements
        for child in self.children:
            if not child.attachable():
                raise ValueError('child is not attachable')
            if child.shape != shape:
                raise ValueError('inconsistent shapes')
            if child.dtype != dtype:
                raise TypeError('inconsistent dtypes')
            for dimension in child.dimensions:
                if dimension.frozen:
                    raise ValueError('child has frozen dimensions')
        for child in self.children:
            for x, y in zip(dimensions, child.dimensions):
                x.link(y)
        super().__init__(dimensions, dtype)
        for child in self.children:
            child.is_root = False

    def _bind(self, buffer: Optional[DeviceArray]) -> None:
        super()._bind(buffer)
        for child in self.children:
            child._bind(buffer)


class AliasIOSlot(IOSlotBase):
    """Slot that allocates a single low-level buffer to back multiple children.

    The child slots need not have the same type or shape. This is typically
    used when logically distinct slots can share memory because they do not
    contain live data at the same time.

    Parameters
    ----------
    children

    Raises
    ------
    ValueError
        If `children` is empty or contains non-attachable elements
    """

    def __init__(self, children: Iterable[IOSlotBase]) -> None:
        super().__init__()
        self.children = list(children)
        self.raw = None       # type: Optional[Any]
        if not len(self.children):
            raise ValueError('empty child list')
        for child in self.children:
            if not child.attachable():
                raise ValueError('child is not attachable')
        for child in self.children:
            child.is_root = False

    def is_bound(self) -> bool:
        return self.raw is not None

    def required_bytes(self) -> int:
        return max([child.required_bytes() for child in self.children])

    def _allocate_host(self, context: AbstractContext) -> HostArray:
        return HostArray((self.required_bytes(),), np.uint8, context=context)

    def _allocate(self, allocator: AbstractAllocator[_RB],
                  raw: Optional[_RB] = None, *, bind: bool = True) -> _RB:
        if raw is None:
            raw = allocator.allocate_raw(self.required_bytes())
        if bind:
            for child in self.children:
                child._allocate(allocator, raw, bind=bind)
            self.raw = raw
        return raw


class Operation(ABC):
    """An instance of a device operation.

    Typically one first creates a template (which contains the program code,
    and is expensive to create) and then instantiates it for use with specific
    buffers.

    An instance of this class contains *slots*, which are named instances of
    :class:`IOSlot`. The user binds specific buffers to these slots to specify
    the memory used in the operation.

    This class is only useful when subclassed. The subclass will populate
    the slots. Subclasses also provide a `_run` function that handles the
    implementation of `__call__`.

    Operations are arranged in a tree, with internal nodes subclassing
    :class:`OperationSequence`. Internal nodes provide slots that proxy for
    the slots of their children; the children's slots should not be
    manipulated directly. When a parent ceases to exist, its children should
    no longer be used i.e., they cannot be reattached to a new parent.

    Parameters
    ----------
    command_queue
        Command queue for the operation
    allocator
        Allocator used to allocate unbound slots

    Attributes
    ----------
    slots : dictionary
        Maps names to slot instances
    hidden_slots : dictionary
        Extra slots which are not root slots, and hence cannot have buffers
        bound to them, but whose buffers can still be referenced. This is
        generally used by :class:`OperationSequence` when several slots are
        aliased to the same memory.
    is_root : boolean
        True if this operation is not part of an :class:`OperationSequence`.
    """

    def __init__(self, command_queue: AbstractCommandQueue,
                 allocator: Optional[AbstractAllocator] = None) -> None:
        if allocator is None:
            allocator = DeviceAllocator(command_queue.context)
        elif allocator.context is not None:
            if allocator.context is not command_queue.context:
                raise ValueError('command_queue and allocator have different contexts')
        self.slots = {}            # type: Dict[str, IOSlotBase]
        self.hidden_slots = {}     # type: Dict[str, IOSlotBase]
        self.command_queue = command_queue
        self.is_root = True
        self.allocator = allocator

    def bind(self, **kwargs) -> None:
        """Bind buffers to slots by keyword.

        Each keyword argument name specifies a slot name.

        Raises
        ------
        KeyError
            if a named slots does not exist
        TypeError
            if a named slot if an alias slot
        """
        for name, buffer in kwargs.items():
            slot = self.slots[name]
            if not isinstance(slot, IOSlot):
                raise TypeError(f'Slot {slot} is not an IOSlot')
            slot.bind(buffer)

    def ensure_bound(self, name: str) -> None:
        """Make sure that a specific slot has a buffer bound, allocating if necessary."""
        slot = self.slots[name]
        if not slot.is_bound():
            slot.allocate(self.allocator)

    def ensure_all_bound(self) -> None:
        """Make sure that all slots have a buffer bound, allocating if necessary."""
        for slot in self.slots.values():
            if not slot.is_bound():
                slot.allocate(self.allocator)

    def buffer(self, name: str) -> DeviceArray:
        """Retrieve the buffer bound to a slot.

        It will consult both :attr:`slots` and :attr:`hidden_slots`.

        Parameters
        ----------
        name : str
            Name of the slot to access

        Returns
        -------
        :class:`DeviceArray`
            Buffer bound to slot `name`.

        Raises
        ------
        KeyError
            If no slot with this name exists
        TypeError
            If the slot exists but it is an alias slot
        ValueError
            If the slot exists but does not yet have a buffer bound
        """
        try:
            slot = self.slots[name]
        except KeyError:
            try:
                slot = self.hidden_slots[name]
            except KeyError:
                raise KeyError('no slot named ' + name) from None
        if isinstance(slot, IOSlot):
            if slot.buffer is None:
                raise ValueError('slot ' + name + ' has no buffer bound')
            return slot.buffer
        else:
            raise TypeError('slot ' + name + ' is an alias slot')

    def required_bytes(self) -> int:
        """Return number of bytes of device storage required."""
        return sum([x.required_bytes() for x in self.slots.values()])

    def parameters(self) -> Mapping[str, Any]:
        """Return dictionary of configuration options for this operation."""
        return {}

    @abstractmethod
    def _run(self) -> Any:
        raise NotImplementedError('abstract base class')

    def __call__(self, **kwargs) -> Any:
        """Run the operation.

        Any slots that are not already bound will allocate new buffers.
        Keyword arguments are passed to :meth:`bind`.
        """
        self.bind(**kwargs)
        self.ensure_all_bound()
        return self._run()


class OperationSequence(Operation):
    """Convenience class for an operation that is built up of smaller named operations.

    Slots are mapped to share data between these smaller operations.
    Initially, each slot named *slot* in a child named *op* is remapped to a
    parent slot named *op*:*slot*. After this, each set provided in `compounds`
    is removed from the slots and combined into a single compound slot.
    Finally, each set in `aliases` is removed and combined into an alias slot.

    For both `compounds` and `aliases`, if a child does not exist, it is
    skipped, and if none of the children exist, no action is taken.

    Parameters
    ----------
    command_queue
        Command queue for the operation
    operations : sequence of 2-tuples
        Name, operation pairs to add. Calling the operation executes them in order
    compounds : mapping of `str` to sequence of `str`, optional
        Names for compound slots, mapped to the original slot names that are replaced
    aliases : mapping of `str` to sequence of `str`, optional
        Names for alias slots, mapped to the original slot names that are replaced
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """

    def __init__(self, command_queue: AbstractCommandQueue,
                 operations: Iterable[Tuple[str, Operation]],
                 compounds: Optional[Mapping[str, Iterable[str]]] = None,
                 aliases: Optional[Mapping[str, Iterable[str]]] = None,
                 allocator: Optional[AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        self.operations = OrderedDict(operations)
        for (name, operation) in self.operations.items():
            if operation.command_queue is not command_queue:
                raise ValueError('child has a different command queue to the parent')
            if not operation.is_root:
                raise ValueError('child already has another parent')
            for (slot_name, slot) in operation.slots.items():
                self.slots[name + ':' + slot_name] = slot
        if compounds is not None:
            for (name, child_names) in compounds.items():
                children = self._extract_slots(child_names, False)
                if children:
                    for child in children:
                        if not isinstance(child, IOSlot):
                            raise TypeError(f'Children of {name} must all be IOSlots')
                    self.slots[name] = CompoundIOSlot(cast(Iterable[IOSlot], children))
        if aliases is not None:
            for (name, child_names) in aliases.items():
                children = self._extract_slots(child_names, True)
                if children:
                    self.slots[name] = AliasIOSlot(children)
        for operation in self.operations.values():
            operation.is_root = False

    def _extract_slots(self, names: Iterable[str], add_to_hidden: bool) -> List[IOSlotBase]:
        """Remove and return the slots with the given names."""
        ans = []      # type: List[IOSlotBase]
        for name in names:
            try:
                ans.append(self.slots[name])
                del self.slots[name]
                if add_to_hidden:
                    assert name not in self.hidden_slots
                    self.hidden_slots[name] = ans[-1]
            except KeyError:
                # Ignore missing items
                pass
        return ans

    def _run(self) -> None:
        for operation in self.operations.values():
            operation()


def _slot_label(slot: IOSlotBase, name: str, hide_detail: bool) -> str:
    """Produce contents of HTML label for a slot.

    If `hide_detail` is true, it will not show the shapes and dtype. This is used when
    the details will appear on a parent.
    """
    if isinstance(slot, AliasIOSlot):
        return f'<b>{name}</b><br/><font color="red">alias slot</font>'
    elif hide_detail:
        return f'<b>{name}</b>'
    else:
        assert isinstance(slot, IOSlot)
        shape = '×'.join(str(x) for x in slot.shape)
        if not shape:
            shape = 'scalar'
        padded_shape = '×'.join(str(x) for x in slot.required_padded_shape())
        dtype = str(np.dtype(slot.dtype))
        return '<b>{name}</b><br/>{shape}<br/>{padded_shape}<br/>{dtype}'.format(
            name=name, shape=shape, padded_shape=padded_shape, dtype=dtype)


def _visualize_operation(
        g: 'graphviz.Graph',
        operation: Operation,
        path: Tuple[str, ...],
        slot_map: MutableMapping[IOSlotBase, str],
        no_detail_slots: Set[IOSlotBase],
        edges: Set[Tuple[IOSlotBase, IOSlotBase]]) -> None:
    """Recursive implementation of :func:`visualize_operation`.

    Parameters
    ----------
    g
        Graph or subgraph to be updated with the operation. If the operation has
        children, it will be rendered as a subgraph, otherwise as a single slot
        with ports for the slots.
    operation
        Operation to add to the graph
    path
        Hierarchical name of `operation`. The last component is used for visual
        display, and the rest for generating unique node labels.
    slot_map
        Mapping from slots to node names (or port names). It is updated in place
        with all the slots in this operation and its descendants.
    no_detail_slots
        Slots that should not show shape and dtype details. It is updated in
        place with any descendant slots of this node.
    edges
        Edges to draw, as tuples of two :class:`IOSlotBase`s.
    """
    all_slots = {**operation.slots, **operation.hidden_slots}

    if isinstance(operation, OperationSequence):
        # Graphviz treats subgraphs with the prefix 'cluster_' specially. We
        # only want this for real subgraphs, not the root.
        node_name = 'cluster_' + '!'.join(path) if len(path) > 1 else 'root'
        for slot_name, slot in all_slots.items():
            if isinstance(slot, CompoundIOSlot):
                for child_slot in slot.children:
                    no_detail_slots.add(child_slot)
        with g.subgraph(name=node_name) as sub:
            sub.attr('graph', label=path[-1])
            for child_name, child in operation.operations.items():
                _visualize_operation(sub, child, path + (child_name,),
                                     slot_map, no_detail_slots, edges)
            for slot_name, slot in list(all_slots.items()):
                if slot not in slot_map:     # It's already represented in the child
                    # ":" has special meaning to graphviz, so we replace it with "+".
                    full_name = 'slot_{}_{}'.format('!'.join(path), slot_name.replace(':', '+'))
                    slot_map[slot] = full_name
                    slot_label = '<' + _slot_label(slot, slot_name, slot in no_detail_slots) + '>'
                    sub.node(full_name, label=slot_label, shape='box')
                else:
                    del all_slots[slot_name]
    else:
        node_name = 'op_' + '!'.join(path)
        label = io.StringIO()
        label.write('<<table border="1" cellborder="0" rows="*" columns="*">'
                    '<tr><td colspan="{n_slots}">{name}</td></tr><tr>'
                    .format(name=path[-1], n_slots=len(operation.slots)))
        for slot_name, slot in operation.slots.items():
            label.write('<td port="{name}"><font point-size="9"><i>{name}</i></font></td>'
                        .format(name=slot_name))
            slot_map[slot] = node_name + ':' + slot_name.replace(':', '+')
        label.write('</tr></table>>')
        g.node(node_name, label=label.getvalue(), shape='plain')

    for slot in all_slots.values():
        if hasattr(slot, 'children'):
            children = getattr(slot, 'children')  # type: Sequence[IOSlotBase]
            for child2 in children:
                edges.add((child2, slot))


def visualize_operation(operation: Operation, filename: str) -> None:
    """Write a visualization of an :class:`Operation` to file.

    This requires the `graphviz package`_ to be installed.

    .. _graphviz package: https://graphviz.readthedocs.io/en/stable/

    Parameters
    ----------
    operation
        Operation or operation sequence to visualize
    filename
        Base filename to write. It should end in ``.gv``, and a rendered
        PDF will automatically be produced with a filename ending in
        ``.gv.pdf``.
    """
    from graphviz import Graph

    g = Graph()
    g.attr('graph', clusterrank='local', ranksep='3', compound='yes')
    slot_map = {}        # type: Dict[IOSlotBase, str]
    edges = set()        # type: Set[Tuple[IOSlotBase, IOSlotBase]]
    _visualize_operation(g, operation, ('root',), slot_map, set(), edges)
    for a, b in edges:
        g.edge(slot_map[a], slot_map[b])
    g.render(filename)



"""Utilities for testing with nosetests.

These are maintained for backwards compatibility. New code is encouraged to use
pytest instead.
"""

import functools
import inspect
import sys
from unittest import mock
from typing import Tuple, Optional, Callable, Awaitable, TypeVar

from decorator import decorator
# Don't import nose here, to allow docs to build even without nose installed.

from katsdpsigproc import accel, tune
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue

_T = TypeVar('_T')
_F = TypeVar('_F', bound=Callable)
_test_context = None           # type: Optional[AbstractContext]
_test_command_queue = None     # type: Optional[AbstractCommandQueue]
_test_initialized = False


def _prepare_device_test() -> Tuple[AbstractContext, AbstractCommandQueue]:
    from nose.plugins.skip import SkipTest

    global _test_initialized, _test_context, _test_command_queue
    if not _test_initialized:
        try:
            _test_context = accel.create_some_context(False)
            _test_command_queue = _test_context.create_command_queue()
            print("Testing on {} ({})".format(
                _test_context.device.name, _test_context.device.platform_name),
                file=sys.stderr)
        except RuntimeError:
            pass  # No devices available
        _test_initialized = True

    if not _test_context:
        raise SkipTest('CUDA/OpenCL not found')
    assert _test_command_queue is not None
    return _test_context, _test_command_queue


def _device_test_sync(test: Callable[..., _T]) -> Callable[..., _T]:
    @functools.wraps(test)
    def wrapper(*args, **kwargs) -> _T:
        context, command_queue = _prepare_device_test()
        with mock.patch('katsdpsigproc.tune.autotuner_impl', new=tune.stub_autotuner):
            args += (context, command_queue)
            # Make the context current (for CUDA contexts). Ideally the test
            # should not depend on this, but PyCUDA leaks memory if objects
            # are deleted without the context current.
            with context:
                return test(*args, **kwargs)
    return wrapper


def _device_test_async(test: Callable[..., Awaitable[_T]]) -> Callable[..., Awaitable[_T]]:
    @functools.wraps(test)
    async def wrapper(*args, **kwargs) -> _T:
        context, command_queue = _prepare_device_test()
        with mock.patch('katsdpsigproc.tune.autotuner_impl', new=tune.stub_autotuner):
            args += (context, command_queue)
            with context:
                return await test(*args, **kwargs)
    return wrapper


def device_test(test: Callable[..., _T]) -> Callable[..., _T]:
    """Decorate an on-device test.

    It provides a context and command queue to the test, skipping it if a
    compute device is not available. It also disables autotuning, instead
    using the `test` value provided for the autotune function.

    If autotuning is desired, use :func:`force_autotune` inside (hence,
    afterwards on the decorator list) this one.
    """
    if inspect.iscoroutinefunction(test):
        return _device_test_async(test)    # type: ignore
    else:
        return _device_test_sync(test)


def cuda_test(test: _F) -> _F:
    """Skip a test if the device is not a CUDA device.

    Put this *after* :meth:`device_test`.
    """
    from nose.plugins.skip import SkipTest

    @functools.wraps(test)
    def wrapper(*args, **kwargs):
        global _test_context
        if not _test_context.device.is_cuda:
            raise SkipTest('Device is not a CUDA device')
        return test(*args, **kwargs)
    return wrapper     # type: ignore


@decorator
def force_autotune(test: Callable[..., _T], *args, **kw) -> _T:
    """Force autotuning for a test (decorator).

    It bypasses the autotuning cache so that the autotuning code always runs.
    """
    with mock.patch('katsdpsigproc.tune.autotuner_impl', new=tune.force_autotuner):
        return test(*args, **kw)


# Prevent nose from treating it as a test
device_test.__test__ = False         # type: ignore
cuda_test.__test__ = False           # type: ignore



#!/usr/bin/env python
from katsdpsigproc import accel, transpose
import numpy as np

dtype = np.complex64
ctype = 'float2'
# dtype = np.float32
# ctype = 'float'

R = 3072
C = 8320
ctx = accel.create_some_context(True)
template = transpose.TransposeTemplate(ctx, dtype, ctype)

queue = ctx.create_tuning_command_queue()
proc = template.instantiate(queue, (R, C))
proc.ensure_all_bound()
rm = proc.buffer('src')
cm = proc.buffer('dest')
proc()  # Warmup
for i in range(4):
    queue.start_tuning()
    proc()
    print(queue.stop_tuning())



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.percentile`."""

from typing import Tuple, Optional, cast

import numpy as np
import pytest

from . import complex_normal
from katsdpsigproc import accel
from katsdpsigproc import percentile
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue


class TestPercentile5:
    @classmethod
    def pad_dimension(cls, dim: accel.Dimension, extra: int) -> None:
        """Modify `dim` to have at least `extra` padding."""
        newdim = accel.Dimension(dim.size, min_padded_size=dim.size + extra)
        newdim.link(dim)

    @pytest.mark.parametrize(
        'R, C, is_amplitude, column_range',
        [
            (4096, 1, False, None),
            (4095, 4029, True, (0, 4009)),
            (4094, 4030, False, (100, 4030)),
            (2343, 6031, False, (123, 4001)),
            (4092, 4032, True, None)
        ]
    )
    def test_percentile5(self, R: int, C: int, is_amplitude: bool,
                         column_range: Optional[Tuple[int, int]],
                         context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        template = percentile.Percentile5Template(context, max_columns=5000,
                                                  is_amplitude=is_amplitude)
        fn = template.instantiate(command_queue, (R, C), column_range)
        # Force some padded, to check that stride calculation works
        src_slot = cast(accel.IOSlot, fn.slots['src'])
        dest_slot = cast(accel.IOSlot, fn.slots['dest'])
        self.pad_dimension(src_slot.dimensions[0], 1)
        self.pad_dimension(src_slot.dimensions[1], 4)
        self.pad_dimension(dest_slot.dimensions[0], 2)
        self.pad_dimension(dest_slot.dimensions[1], 3)
        rs = np.random.RandomState(seed=1)
        if is_amplitude:
            ary = np.abs(rs.randn(R, C)).astype(np.float32)  # note positive numbers required
        else:
            ary = complex_normal(rs, size=(R, C)).astype(np.complex64)
        src = fn.slots['src'].allocate(fn.allocator)
        dest = fn.slots['dest'].allocate(fn.allocator)
        src.set_async(command_queue, ary)
        fn()
        out = dest.get(command_queue)
        if column_range is None:
            column_range = (0, C)
        expected = np.percentile(np.abs(ary[:, column_range[0]:column_range[1]]),
                                 [0, 100, 25, 75, 50], axis=1, interpolation='lower')
        expected = expected.astype(dtype=np.float32)
        # When amplitudes are being computed, we won't get a bit-exact match
        if is_amplitude:
            np.testing.assert_equal(expected, out)
        else:
            np.testing.assert_allclose(expected, out, 1e-6)

    @pytest.mark.force_autotune
    def test_autotune(self, context: AbstractContext) -> None:
        """Check that the autotuner runs successfully."""
        percentile.Percentile5Template(context, max_columns=5000)



import numpy as np
import katsdpsigproc.accel

SOURCE = """
__kernel void multiply(__global float *data, float scale) {
    data[get_global_id(0)] *= scale;
}
"""

ctx = katsdpsigproc.accel.create_some_context(device_filter=lambda x: not x.is_cuda)
queue = ctx.create_command_queue()
buf = katsdpsigproc.accel.DeviceArray(ctx, (50,), np.float32, (64,))
host = buf.empty_like()
host[:] = np.random.uniform(size=host.shape)
buf.set(queue, host)

program = ctx.compile(SOURCE)
kernel = program.get_kernel('multiply')
queue.enqueue_kernel(
    kernel,
    [buf.buffer, np.float32(3.0)],
    global_size=buf.padded_shape,
    local_size=(32,)
)
buf.get(queue, host)
print(host)



import numpy as np
import katsdpsigproc.accel
import katsdpsigproc.fill
import katsdpsigproc.reduce


class FillReduceTemplate:
    def __init__(self, context):
        self.fill = katsdpsigproc.fill.FillTemplate(context, np.float32, 'float')
        self.hreduce = katsdpsigproc.reduce.HReduceTemplate(context, np.float32, 'float', 'a+b', '0.0f')

    def instantiate(self, queue, shape):
        return FillReduce(self, queue, shape)


class FillReduce(katsdpsigproc.accel.OperationSequence):
    def __init__(self, template, queue, shape):
        self.fill = template.fill.instantiate(queue, shape)
        self.hreduce = template.hreduce.instantiate(queue, shape)
        operations = [
            ('fill', self.fill),
            ('hreduce', self.hreduce)
        ]
        compounds = {
            'src': ['fill:data', 'hreduce:src'],
            'dest': ['hreduce:dest']
        }
        super().__init__(queue, operations, compounds)
        self.template = template

    def __call__(self, fill_value):
        self.fill.set_value(fill_value)
        super().__call__()


ctx = katsdpsigproc.accel.create_some_context()
queue = ctx.create_command_queue()
op_template = FillReduceTemplate(ctx)
op = op_template.instantiate(queue, (10, 5))
op(42)
print(op.buffer('dest').get(queue))



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Perform on-device percentile calculation of 2D arrays."""
# see scripts/percentiletest.py for an example

from typing import Tuple, Mapping, Callable, Optional, Any, cast
from typing_extensions import TypedDict

import numpy as np

from . import accel
from . import tune
from .abc import AbstractContext, AbstractCommandQueue


_TuningDict = TypedDict('_TuningDict', {'size': int, 'wgsy': int})


class Percentile5Template:
    """Kernel for calculating percentiles of a 2D array of data.

    5 percentiles [0,100,25,75,50] are calculated per row (along columns, independently per row).
    The lower percentile element, rather than a linear interpolation is chosen.
    WARNING: assumes all values are positive.

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    max_columns
        Maximum number of columns
    is_amplitude
        If true, the inputs are scalar amplitudes; if false, they are complex
        numbers and the answers are computed on the absolute values
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - size: number of workitems per workgroup along each row
        - wgsy: number of workitems per workgroup along each column
    """

    autotune_version = 8

    def __init__(self, context: AbstractContext, max_columns: int,
                 is_amplitude: bool = True, tuning: Optional[_TuningDict] = None) -> None:
        self.context = context
        self.max_columns = max_columns
        self.is_amplitude = is_amplitude

        if tuning is None:
            tuning = self.autotune(context, max_columns, is_amplitude)
        self.size = tuning['size']
        self.wgsy = tuning['wgsy']
        self.vt = accel.divup(max_columns, tuning['size'])
        self.program = accel.build(context, "percentile.mako", {
            'size': self.size,
            'wgsy': self.wgsy,
            'vt': self.vt,
            'is_amplitude': self.is_amplitude
        })

    @classmethod
    @tune.autotuner(test={'size': 64, 'wgsy': 4})
    def autotune(cls, context: AbstractContext, max_columns: int,
                 is_amplitude: bool) -> _TuningDict:
        queue = context.create_tuning_command_queue()
        in_shape = (4096, max_columns)
        rs = np.random.RandomState(seed=1)
        if is_amplitude:
            host_data: np.ndarray = rs.uniform(size=in_shape).astype(np.float32)
        else:
            host_data = rs.standard_normal(in_shape) + 1j * rs.standard_normal(in_shape)
            host_data = host_data.astype(np.complex64)

        def generate(size: int, wgsy: int) -> Callable[[int], float]:
            if size * wgsy < 32 or size * wgsy > 1024:
                raise RuntimeError('work group size is unnecessarily large or small, skipping')
            if max_columns > size * 256:
                raise RuntimeError('too many columns')
            fn = cls(context, max_columns, is_amplitude, {
                'size': size, 'wgsy': wgsy}).instantiate(queue, in_shape)
            inp = fn.slots['src'].allocate(fn.allocator)
            fn.slots['dest'].allocate(fn.allocator)
            inp.set(queue, host_data)
            return tune.make_measure(queue, fn)

        return cast(_TuningDict, tune.autotune(generate,
                                               size=[8, 16, 32, 64, 128, 256, 512, 1024],
                                               wgsy=[1, 2, 4, 8, 16, 32]))

    def instantiate(self, command_queue: AbstractCommandQueue,
                    shape: Tuple[int, int],
                    column_range: Optional[Tuple[int, int]] = None,
                    allocator: Optional[accel.AbstractAllocator] = None) -> 'Percentile5':
        return Percentile5(self, command_queue, shape, column_range, allocator)


class Percentile5(accel.Operation):
    """Concrete instance of :class:`PercentileTemplate`.

    .. warning::

       Assumes all values are positive when `template.is_amplitude` is `True`.

    .. rubric:: Slots

    **src**
        Input type float32 or complex64.
        Shape is number of rows by number of columns, where 5 percentiles
        are computed along the columns, per row.

    **dest**
        Output type float32.
        Shape is (5, number of rows of input)

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    shape
        Shape of the source data
    column_range:
        Half-open interval of columns that will be processed. If not specified, all columns are
        processed.
    allocator
        Allocator used to allocate unbound slots
    """

    def __init__(self, template: Percentile5Template, command_queue: AbstractCommandQueue,
                 shape: Tuple[int, int], column_range: Optional[Tuple[int, int]],
                 allocator: Optional[accel.AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        if column_range is None:
            column_range = (0, shape[1])
        if column_range[1] <= column_range[0]:
            raise ValueError('column range is empty')
        if column_range[0] < 0 or column_range[1] > shape[1]:
            raise IndexError('column range is out of range')
        if column_range[1] - column_range[0] > template.max_columns:
            raise ValueError('columns exceeds max_columns')
        self.template = template
        self.kernel = template.program.get_kernel("percentile5_float")
        self.shape = shape
        self.column_range = column_range
        src_type = np.float32 if self.template.is_amplitude else np.complex64
        row_dim = accel.Dimension(shape[0], self.template.wgsy)
        col_dim = accel.Dimension(shape[1])
        self.slots['src'] = accel.IOSlot((row_dim, col_dim), src_type)
        self.slots['dest'] = accel.IOSlot((5, row_dim), np.float32)

    def _run(self) -> None:
        src = self.buffer('src')
        dest = self.buffer('dest')
        rows_padded = accel.roundup(src.shape[0], self.template.wgsy)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [
                src.buffer, dest.buffer,
                np.int32(src.padded_shape[1]),
                np.int32(dest.padded_shape[1]),
                np.int32(self.column_range[0]),
                np.int32(self.column_range[1] - self.column_range[0])
            ],
            global_size=(self.template.size, rows_padded),
            local_size=(self.template.size, self.template.wgsy))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'max_columns': self.template.max_columns,
            'is_amplitude': self.template.is_amplitude,
            'shape': self.slots['src'].shape,      # type: ignore
            'column_range': self.column_range
        }



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Fill device array with a constant value."""

from typing import Tuple, Mapping, Callable, Optional, Any

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any     # type: ignore

from . import accel
from . import tune
from .abc import AbstractContext, AbstractCommandQueue


class FillTemplate:
    """Fill a device array with a constant value.

    The pad elements are also filled with this value.

    .. note::
        To fill with zeros, use :meth:`.DeviceArray.zero`

    Parameters
    ----------
    context
        Context for which kernels will be compiled
    dtype
        Type of data elements
    ctype
        Type (in C/CUDA, not numpy) of data elements
    tuning
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgs: number of workitems per workgroup
    """

    def __init__(self, context: AbstractContext, dtype: DTypeLike, ctype: str,
                 tuning: Optional[Mapping[str, Any]] = None) -> None:
        self.context = context
        self.dtype: np.dtype = np.dtype(dtype)
        self.ctype = ctype
        if tuning is None:
            tuning = self.autotune(context, dtype, ctype)
        self.wgs = tuning['wgs']
        self.program = accel.build(context, "fill.mako", {
            'wgs': self.wgs,
            'ctype': ctype
        })

    @classmethod
    @tune.autotuner(test={'wgs': 128})
    def autotune(cls, context: AbstractContext, dtype: DTypeLike, ctype: str) -> Mapping[str, Any]:
        queue = context.create_tuning_command_queue()
        shape = (1048576,)
        data = accel.DeviceArray(context, shape, dtype=dtype)

        def generate(wgs: int) -> Callable[[int], float]:
            fn = cls(context, dtype, ctype, {'wgs': wgs}).instantiate(queue, shape)
            fn.bind(data=data)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate, wgs=[64, 128, 256, 512])

    def instantiate(self, command_queue: AbstractCommandQueue, shape: Tuple[int, ...],
                    allocator: Optional[accel.AbstractAllocator] = None) -> 'Fill':
        return Fill(self, command_queue, shape, allocator)


class Fill(accel.Operation):
    """Concrete instance of :class:`FillTemplate`.

    .. rubric:: Slots

    **data**
        Array to be filled (padding will be filled too)

    Parameters
    ----------
    template
        Operation template
    command_queue
        Command queue for the operation
    shape
        Shape for the data slot
    allocator
        Allocator used to allocate unbound slots
    """

    def __init__(self, template: FillTemplate, command_queue: AbstractCommandQueue,
                 shape: Tuple[int, ...],
                 allocator: Optional[accel.AbstractAllocator] = None) -> None:
        super().__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel("fill")
        self.shape = shape
        self.slots['data'] = accel.IOSlot(shape, self.template.dtype)
        self.value = self.template.dtype.type()

    def set_value(self, value: Any) -> None:
        self.value = self.template.dtype.type(value)

    def _run(self) -> None:
        data = self.buffer('data')

        elements = int(np.prod(data.padded_shape))
        global_size = accel.roundup(elements, self.template.wgs)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [data.buffer, np.uint32(elements), self.value],
            global_size=(global_size,),
            local_size=(self.template.wgs,))

    def parameters(self) -> Mapping[str, Any]:
        return {
            'dtype': self.template.dtype,
            'ctype': self.template.ctype,
            'shape': self.slots['data'].shape,     # type: ignore
            'value': self.value
        }



################################################################################
# Copyright (c) 2014-2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Abstraction layer over PyCUDA.

It implements the abstract interfaces defined in :mod:`katsdpsigproc.abc`.
"""

from typing import List, Tuple, Sequence, Optional, Type, TypeVar, Union, Any
from types import TracebackType

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any     # type: ignore
import pycuda.driver
import pycuda.compiler
import pycuda.gpuarray
import pycuda.characterize

from .abc import (AbstractProgram, AbstractKernel, AbstractDevice, AbstractContext,
                  AbstractEvent, AbstractCommandQueue, AbstractTuningCommandQueue)


# When building with autodoc_mock_imports, DEFAULT_NVCC_FLAGS is a mock that
# doesn't have a + operator.
if isinstance(pycuda.compiler.DEFAULT_NVCC_FLAGS, list):
    NVCC_FLAGS = pycuda.compiler.DEFAULT_NVCC_FLAGS + ['-lineinfo']
else:
    NVCC_FLAGS = ['-lineinfo']
_T = TypeVar('_T')
_C = TypeVar('_C', bound='Context')
_D = TypeVar('_D', bound='Device')
_AnyBuffer = Union[pycuda.gpuarray.GPUArray, np.ndarray]


class Program(AbstractProgram['Kernel']):
    def __init__(self, pycuda_module: pycuda.driver.Module) -> None:
        self._pycuda_program = pycuda_module

    def get_kernel(self, name: str) -> 'Kernel':
        return Kernel(self, name)


class Kernel(AbstractKernel[Program]):
    def __init__(self, program: Program, name: str) -> None:
        self._pycuda_kernel = program._pycuda_program.get_function(name)


class Event(AbstractEvent):
    def __init__(self, pycuda_event: pycuda.driver.Event) -> None:
        self._pycuda_event = pycuda_event

    def wait(self) -> None:
        self._pycuda_event.synchronize()

    def time_since(self, prior_event: 'Event') -> float:
        prior_event.wait()
        self.wait()
        return 1e-3 * self._pycuda_event.time_since(prior_event._pycuda_event)

    def time_till(self, next_event: 'Event') -> float:
        return next_event.time_since(self)


class Device(AbstractDevice['Context']):
    def __init__(self, pycuda_device: pycuda.driver.Device) -> None:
        self._pycuda_device = pycuda_device

    def make_context(self) -> 'Context':
        pycuda_context = self._pycuda_device.make_context()
        # The above also makes the context current, which we do not
        # want (it leads to errors on termination).
        pycuda_context.pop()
        return Context(pycuda_context)

    @property
    def name(self) -> str:
        return self._pycuda_device.name()

    @property
    def platform_name(self) -> str:
        return 'CUDA'

    @property
    def driver_version(self) -> str:
        return 'CUDA:{0[0]}{0[1]}{0[2]} Driver:{1}'.format(
            pycuda.driver.get_version(), pycuda.driver.get_driver_version())

    @property
    def is_cuda(self) -> bool:
        return True

    @property
    def is_gpu(self) -> bool:
        return True

    @property
    def is_accelerator(self) -> bool:
        return False

    @property
    def is_cpu(self) -> bool:
        return False

    @property
    def simd_group_size(self) -> int:
        return self._pycuda_device.warp_size

    @property
    def compute_capability(self) -> Tuple[int, int]:
        return self._pycuda_device.compute_capability()

    @classmethod
    def get_devices(cls: Type[_D]) -> Sequence[_D]:
        num_devices = pycuda.driver.Device.count()
        return [cls(pycuda.driver.Device(i)) for i in range(num_devices)]

    @classmethod
    def get_devices_by_platform(cls: Type[_D]) -> Sequence[Sequence[_D]]:
        return [cls.get_devices()]


class _RawManaged:
    """Wrap a PyCUDA managed allocation into an opaque object.

    The managed memory API is different to the device memory API, in that one
    cannot allocate a raw pointer. Thus, "raw" allocations are wrapped in this
    class so that they are not accidentally used as numpy arrays.
    """

    def __init__(self, wrapped: np.ndarray) -> None:
        self._wrapped = wrapped

    def get_array(self, shape: Tuple[int, ...], dtype: DTypeLike) -> np.ndarray:
        """Return a view of (a prefix of) the memory, with the given shape and type."""
        size = int(np.prod(shape)) * np.dtype(dtype).itemsize
        return self._wrapped[:size].view(dtype).reshape(shape)


class Context(AbstractContext[pycuda.gpuarray.GPUArray,
                              pycuda.driver.DeviceAllocation,
                              _RawManaged,
                              Device, Program, 'CommandQueue', 'TuningCommandQueue']):
    def __init__(self, pycuda_context: pycuda.driver.Context) -> None:
        self._pycuda_context = pycuda_context

    @property
    def device(self) -> Device:
        with self:
            return Device(pycuda.driver.Context.get_device())

    def compile(self, source: str, extra_flags: Optional[List[str]] = None) -> Program:
        if extra_flags is None:
            extra_flags = []
        with self:
            module = pycuda.compiler.SourceModule(source, options=NVCC_FLAGS + extra_flags)
            return Program(module)

    def allocate_raw(self, n_bytes: int) -> pycuda.driver.DeviceAllocation:
        with self:
            return pycuda.driver.mem_alloc(n_bytes)

    def allocate(self, shape: Tuple[int, ...], dtype: DTypeLike,
                 raw: Optional[pycuda.driver.DeviceAllocation] = None) -> pycuda.gpuarray.GPUArray:
        with self:
            return pycuda.gpuarray.GPUArray(shape, dtype, gpudata=raw)

    def allocate_pinned(self, shape: Tuple[int, ...], dtype: DTypeLike) -> np.ndarray:
        with self:
            return pycuda.driver.pagelocked_empty(shape, dtype)

    def allocate_svm_raw(self, n_bytes: int) -> _RawManaged:
        with self:
            return _RawManaged(pycuda.driver.managed_empty(
                (n_bytes,), np.uint8, mem_flags=pycuda.driver.mem_attach_flags.GLOBAL))

    def allocate_svm(self, shape: Tuple[int, ...], dtype: DTypeLike,
                     raw: Optional[_RawManaged] = None) -> np.ndarray:
        with self:
            if raw is None:
                return pycuda.driver.managed_empty(
                    shape, dtype, mem_flags=pycuda.driver.mem_attach_flags.GLOBAL)
            else:
                return raw.get_array(shape, dtype)

    def create_command_queue(self, profile: bool = False) -> 'CommandQueue':
        return CommandQueue(self, profile=profile)

    def create_tuning_command_queue(self) -> 'TuningCommandQueue':
        return TuningCommandQueue(self)

    def __enter__(self: _C) -> _C:
        self._pycuda_context.push()
        return self

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        self._pycuda_context.pop()


class CommandQueue(AbstractCommandQueue[pycuda.gpuarray.GPUArray, Context, Event, Kernel]):
    def __init__(self, context: Context, pycuda_stream: Optional[pycuda.driver.Stream] = None,
                 profile: bool = False) -> None:
        self.context = context
        if pycuda_stream is None:
            with context:
                self._pycuda_stream = pycuda.driver.Stream()
        else:
            self._pycuda_stream = pycuda_stream

    def enqueue_read_buffer(self, buffer: pycuda.gpuarray.GPUArray, data: Any,
                            blocking: bool = True) -> None:
        with self.context:
            # CUDA doesn't support synchronous transfers sequenced in a stream
            # (PyCUDA simply doesn't pass on the stream argument), so use an
            # async transfer and then block if necessary.
            buffer.get_async(self._pycuda_stream, data)
            if blocking:
                self._pycuda_stream.synchronize()

    def enqueue_write_buffer(self, buffer: pycuda.gpuarray.GPUArray, data: Any,
                             blocking: bool = True) -> None:
        with self.context:
            # See comment in enqueue_read_buffer
            buffer.set_async(data, self._pycuda_stream)
            if blocking:
                self._pycuda_stream.synchronize()

    @staticmethod
    def _get_device_pointer(buffer: _AnyBuffer) -> int:
        """Retrieve the device pointer from either a GPUArray or a managed memory allocation."""
        if isinstance(buffer, pycuda.gpuarray.GPUArray):
            return int(buffer.gpudata)
        else:
            return buffer.base.get_device_pointer()  # type: ignore

    def enqueue_copy_buffer_rect(
            self, src_buffer: _AnyBuffer,
            dest_buffer: _AnyBuffer,
            src_origin: int, dest_origin: int,
            shape: Sequence[int], src_strides: Sequence[int], dest_strides: Sequence[int]) -> None:
        with self.context:
            assert src_strides[0] == 1
            assert dest_strides[0] == 1
            assert len(shape) > 0 and len(shape) <= 3
            if len(shape) == 1:
                pycuda.driver.memcpy_dtod_async(
                    self._get_device_pointer(dest_buffer) + dest_origin,
                    self._get_device_pointer(src_buffer) + src_origin,
                    shape[0],
                    self._pycuda_stream)
            else:
                if len(shape) == 3:
                    copy = pycuda.driver.Memcpy3D()
                else:
                    copy = pycuda.driver.Memcpy2D()
                copy.src_pitch = src_strides[1]
                copy.set_src_device(self._get_device_pointer(src_buffer) + src_origin)
                copy.dst_pitch = dest_strides[1]
                copy.set_dst_device(self._get_device_pointer(dest_buffer) + dest_origin)
                copy.width_in_bytes = shape[0]
                copy.height = shape[1]
                if len(shape) >= 3:
                    assert src_strides[2] % src_strides[1] == 0
                    assert dest_strides[2] % dest_strides[1] == 0
                    copy.src_height = src_strides[2] // src_strides[1]
                    copy.dst_height = dest_strides[2] // dest_strides[1]
                    copy.depth = shape[2]
                copy(self._pycuda_stream)

    @classmethod
    def _byte_buffer(cls, data: np.ndarray) -> np.ndarray:
        """Reinterpret a contiguous array as an array of bytes."""
        view = data.view()
        view.shape = (data.size,)   # Reshape while disallowing copy
        return view.view(np.uint8)

    def enqueue_read_buffer_rect(
            self, buffer: _AnyBuffer, data: Any,
            buffer_origin: int, data_origin: int, shape: Sequence[int],
            buffer_strides: Sequence[int], data_strides: Sequence[int],
            blocking: bool = True) -> None:
        with self.context:
            assert buffer_strides[0] == 1
            assert data_strides[0] == 1
            assert len(shape) > 0 and len(shape) <= 3
            data = self._byte_buffer(data)
            if len(shape) == 1:
                pycuda.driver.memcpy_dtoh_async(
                    data[data_origin : data_origin + shape[0]],
                    self._get_device_pointer(buffer) + buffer_origin,
                    self._pycuda_stream)
            else:
                if len(shape) == 3:
                    copy = pycuda.driver.Memcpy3D()
                else:
                    copy = pycuda.driver.Memcpy2D()
                copy.src_pitch = buffer_strides[1]
                copy.set_src_device(self._get_device_pointer(buffer) + buffer_origin)
                copy.dst_pitch = data_strides[1]
                copy.set_dst_host(data[data_origin:])
                copy.width_in_bytes = shape[0]
                copy.height = shape[1]
                if len(shape) >= 3:
                    assert data_strides[2] % data_strides[1] == 0
                    assert buffer_strides[2] % buffer_strides[1] == 0
                    copy.src_height = buffer_strides[2] // buffer_strides[1]
                    copy.dst_height = data_strides[2] // data_strides[1]
                    copy.depth = shape[2]
                copy(self._pycuda_stream)
            if blocking:
                self._pycuda_stream.synchronize()

    def enqueue_write_buffer_rect(
            self, buffer: _AnyBuffer, data: Any,
            buffer_origin: int, data_origin: int, shape: Sequence[int],
            buffer_strides: Sequence[int], data_strides: Sequence[int],
            blocking: bool = True) -> None:
        with self.context:
            assert buffer_strides[0] == 1
            assert data_strides[0] == 1
            assert len(shape) > 0 and len(shape) <= 3
            data = self._byte_buffer(data)
            if len(shape) == 1:
                pycuda.driver.memcpy_htod_async(
                    self._get_device_pointer(buffer) + buffer_origin,
                    data[data_origin : data_origin + shape[0]],
                    self._pycuda_stream)
            else:
                if len(shape) == 3:
                    copy = pycuda.driver.Memcpy3D()
                else:
                    copy = pycuda.driver.Memcpy2D()
                copy.src_pitch = data_strides[1]
                copy.set_src_host(data[data_origin:])
                copy.dst_pitch = buffer_strides[1]
                copy.set_dst_device(self._get_device_pointer(buffer) + buffer_origin)
                copy.width_in_bytes = shape[0]
                copy.height = shape[1]
                if len(shape) >= 3:
                    assert data_strides[2] % data_strides[1] == 0
                    assert buffer_strides[2] % buffer_strides[1] == 0
                    copy.src_height = data_strides[2] // data_strides[1]
                    copy.dst_height = buffer_strides[2] // buffer_strides[1]
                    copy.depth = shape[2]
                copy(self._pycuda_stream)
            if blocking:
                self._pycuda_stream.synchronize()

    def enqueue_zero_buffer(self, buffer: _AnyBuffer) -> None:
        with self.context:
            if isinstance(buffer, pycuda.gpuarray.GPUArray):
                pycuda.driver.memset_d8_async(
                    buffer.gpudata, 0, buffer.mem_size * buffer.dtype.itemsize,
                    stream=self._pycuda_stream)
            else:
                # managed memory
                pycuda.driver.memset_d8_async(
                    buffer.base.get_device_pointer(),     # type: ignore
                    0,
                    buffer.size * buffer.dtype.itemsize,
                    stream=self._pycuda_stream)

    def enqueue_kernel(self, kernel: Kernel, args: Sequence[Any],
                       global_size: Tuple[int, ...], local_size: Tuple[int, ...]) -> None:
        assert len(global_size) == len(local_size)
        block = [1, 1, 1]
        grid = [1, 1, 1]
        for i in range(len(global_size)):
            assert global_size[i] % local_size[i] == 0
            block[i] = local_size[i]
            grid[i] = global_size[i] // local_size[i]
        with self.context:
            kernel._pycuda_kernel(*args, block=tuple(block), grid=tuple(grid),
                                  stream=self._pycuda_stream)

    def enqueue_marker(self) -> Event:
        with self.context:
            event = pycuda.driver.Event(pycuda.driver.event_flags.BLOCKING_SYNC)
            event.record(self._pycuda_stream)
        return Event(event)

    def enqueue_wait_for_events(self, events: Sequence[Event]) -> None:
        with self.context:
            for event in events:
                self._pycuda_stream.wait_for_event(event._pycuda_event)

    def flush(self) -> None:
        with self.context:
            # This anecdotally flushes, but isn't tested
            self._pycuda_stream.is_done()

    def finish(self) -> None:
        with self.context:
            self._pycuda_stream.synchronize()


class TuningCommandQueue(CommandQueue,
                         AbstractTuningCommandQueue[pycuda.gpuarray.GPUArray,
                                                    Context, Event, Kernel]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.is_tuning = False
        self._start_event = None    # type: Optional[Event]

    def start_tuning(self) -> None:
        self.is_tuning = True
        self._start_event = self.enqueue_marker()

    def stop_tuning(self) -> float:
        elapsed = 0.0
        if self._start_event is not None:
            end_event = self.enqueue_marker()
            elapsed = self._start_event.time_till(end_event)
        self._start_event = None
        self.is_tuning = False
        return elapsed



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.transpose`."""

from typing import cast

import numpy as np
import pytest

from katsdpsigproc import accel
from katsdpsigproc import transpose
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue


class TestTranspose:
    @classmethod
    def pad_dimension(cls, dim: accel.Dimension, extra: int) -> None:
        """Modify `dim` to have at least `extra` padding."""
        newdim = accel.Dimension(dim.size, min_padded_size=dim.size + extra)
        newdim.link(dim)

    @pytest.mark.parametrize('R, C', [(4, 5), (53, 7), (53, 81), (32, 64)])
    def test_transpose(self, R: int, C: int,
                       context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        template = transpose.TransposeTemplate(context, np.float32, 'float')
        fn = template.instantiate(command_queue, (R, C))
        # Force some padded, to check that stride calculation works
        src_slot = cast(accel.IOSlot, fn.slots['src'])
        dest_slot = cast(accel.IOSlot, fn.slots['dest'])
        self.pad_dimension(src_slot.dimensions[0], 1)
        self.pad_dimension(src_slot.dimensions[1], 4)
        self.pad_dimension(dest_slot.dimensions[0], 2)
        self.pad_dimension(dest_slot.dimensions[1], 3)

        ary = np.random.randn(R, C).astype(np.float32)
        src = fn.slots['src'].allocate(fn.allocator)
        dest = fn.slots['dest'].allocate(fn.allocator)
        src.set_async(command_queue, ary)
        fn()
        out = dest.get(command_queue)
        np.testing.assert_equal(ary.T, out)

    @pytest.mark.force_autotune
    def test_autotune(self, context: AbstractContext) -> None:
        """Check that the autotuner runs successfully."""
        transpose.TransposeTemplate(context, np.uint8, 'unsigned char')



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpsigproc.accel`."""

import shutil
import os
import tempfile
from textwrap import dedent
from unittest import mock
from typing import Tuple, Optional, Callable, Type, Any

import numpy as np
try:
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any     # type: ignore
import pytest
from mako.template import Template

from katsdpsigproc import accel
from katsdpsigproc.accel import HostArray, DeviceArray, SVMArray, LinenoLexer
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue
if accel.have_cuda:
    import pycuda
if accel.have_opencl:
    from katsdpsigproc import opencl
    import pyopencl


class TestLinenoLexer:
    def test_escape_filename(self) -> None:
        assert LinenoLexer._escape_filename(r'abc"def\ghi') == r'"abc\"def\\ghi"'

    def test_render(self) -> None:
        source = "line 1\nline 2\nline 3"
        out = Template(source, lexer_cls=LinenoLexer).render()
        assert out == dedent(
            """\
            #line 1 "<string>"
            line 1
            #line 2 "<string>"
            line 2
            #line 3 "<string>"
            line 3
            """)


class TestHostArray:
    cls = HostArray      # type: Type[HostArray]

    @pytest.fixture
    def factory(self) -> Callable[[Tuple[int, ...], DTypeLike, Tuple[int, ...]], HostArray]:
        def allocate(shape: Tuple[int, ...], dtype: DTypeLike,
                     padded_shape: Tuple[int, ...]) -> HostArray:
            return self.cls(shape, dtype, padded_shape)
        return allocate

    def test_safe(
            self,
            factory: Callable[[Tuple[int, ...], DTypeLike, Tuple[int, ...]], HostArray]) -> None:
        shape = (17, 13)
        padded_shape = (32, 16)
        constructed = factory(shape, np.int32, padded_shape)
        view = np.zeros(padded_shape)[2:4, 3:5].view(self.cls)
        sliced = constructed[2:4, 3:5]
        assert self.cls.safe(constructed)
        assert not self.cls.safe(view)
        assert not self.cls.safe(sliced)
        assert not self.cls.safe(np.zeros(shape))


class TestDeviceArray:
    cls = DeviceArray      # type: Type[DeviceArray]

    @pytest.fixture
    def shape(self) -> Tuple[int, ...]:
        return (17, 13)

    @pytest.fixture
    def padded_shape(self) -> Tuple[int, ...]:
        return (32, 16)

    @pytest.fixture
    def strides(self) -> Tuple[int, ...]:
        return (64, 4)

    @pytest.fixture
    def array(self, context: AbstractContext,
              shape: Tuple[int, ...], padded_shape: Tuple[int, ...]) -> DeviceArray:
        return self.cls(
            context=context,
            shape=shape,
            dtype=np.int32,
            padded_shape=padded_shape)

    def test_properties(self, array: DeviceArray, shape: Tuple[int, ...],
                        padded_shape: Tuple[int, ...], strides: Tuple[int, ...]) -> None:
        assert array.shape == shape
        assert array.padded_shape == padded_shape
        assert array.strides == strides
        assert array.dtype == np.int32

    def test_empty_like(self, array: DeviceArray) -> None:
        ary = array.empty_like()
        assert ary.shape == array.shape
        assert ary.strides == array.strides
        assert HostArray.safe(ary)

    def test_set_get(self, array: DeviceArray,
                     context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        ary = np.random.randint(0, 100, array.shape).astype(np.int32)
        array.set(command_queue, ary)
        # Read back results, check that it matches
        buf = np.zeros(array.padded_shape, dtype=np.int32)
        if isinstance(array, SVMArray):
            buf[:] = array.buffer
        elif context.device.is_cuda:
            with context:
                pycuda.driver.memcpy_dtoh(buf, array.buffer.gpudata)
        else:
            assert isinstance(command_queue, opencl.CommandQueue)
            pyopencl.enqueue_copy(
                command_queue._pyopencl_command_queue,
                buf, array.buffer.data)
        buf = buf[0:array.shape[0], 0:array.shape[1]]
        np.testing.assert_equal(ary, buf)
        # Check that it matches get
        buf = array.get(command_queue)
        np.testing.assert_equal(ary, buf)

    def _test_copy_region(
            self, context: AbstractContext, command_queue: AbstractCommandQueue,
            src_shape: Tuple[int, ...], src_padded_shape: Optional[Tuple[int, ...]],
            dest_shape: Tuple[int, ...], dest_padded_shape: Optional[Tuple[int, ...]],
            src_region: accel._Slice, dest_region: accel._Slice) -> None:
        # copy_region test
        dtype = np.int32
        src = self.cls(context, src_shape, dtype, src_padded_shape)
        dest = self.cls(context, dest_shape, dtype, dest_padded_shape)
        h_src = src.empty_like()
        h_src[...] = np.arange(h_src.size).reshape(src_shape)
        src.set(command_queue, h_src)
        dest.zero(command_queue)
        src.copy_region(command_queue, dest, src_region, dest_region)
        command_queue.finish()   # Needed for the SVMArray test, since the next line is a CPU copy
        h_dest = dest.get(command_queue)
        expected = np.zeros_like(h_dest)
        expected[dest_region] = h_src[src_region]
        np.testing.assert_array_equal(expected, h_dest)
        # set_region test
        dest.zero(command_queue)
        dest.set_region(command_queue, h_src, dest_region, src_region)
        command_queue.finish()
        h_dest = dest.get(command_queue)
        np.testing.assert_array_equal(expected, h_dest)
        # get_region test
        h_dest.fill(0)
        src.get_region(command_queue, h_dest, src_region, dest_region)
        command_queue.finish()
        np.testing.assert_array_equal(expected, h_dest)

    def test_copy_region_4d(self, context: AbstractContext,
                            command_queue: AbstractCommandQueue) -> None:
        self._test_copy_region(
            context, command_queue,
            (7, 5, 12, 19), (8, 9, 14, 25),
            (7, 5, 12, 19), (10, 8, 12, 21),
            np.s_[2:4, :, 3:7:2, 10:], np.s_[1:3, 0:5, 4:8:2, 10:])

    def test_copy_region_0d(self, context: AbstractContext,
                            command_queue: AbstractCommandQueue) -> None:
        self._test_copy_region(context, command_queue, (), (), (), (), (), ())

    def test_copy_region_1d(self, context: AbstractContext,
                            command_queue: AbstractCommandQueue) -> None:
        self._test_copy_region(
            context, command_queue, (3, 5), (4, 6), (5, 7), (7, 10),
            np.s_[1, 0:3], np.s_[2, 1:4])

    def test_copy_region_2d(self, context: AbstractContext,
                            command_queue: AbstractCommandQueue) -> None:
        self._test_copy_region(
            context, command_queue, (3, 5), (4, 6), (5, 7), (7, 10),
            np.s_[:1, 0:3], np.s_[-1:, 1:4])

    def test_copy_region_missing_axes(self, context: AbstractContext,
                                      command_queue: AbstractCommandQueue) -> None:
        self._test_copy_region(
            context, command_queue, (3, 5), (4, 6), (5, 5), (7, 10),
            np.s_[:1], np.s_[-1:])

    def test_copy_region_newaxis(self, context: AbstractContext,
                                 command_queue: AbstractCommandQueue) -> None:
        self._test_copy_region(
            context, command_queue, (10,), None, (10, 10), None,
            np.s_[np.newaxis, 4:6], np.s_[:1, 7:9])

    def test_copy_region_negative(self, context: AbstractContext,
                                  command_queue: AbstractCommandQueue) -> None:
        self._test_copy_region(
            context, command_queue, (10, 12), None, (7, 8), None,
            np.s_[-4, -3:-1], np.s_[-7, -8:-4:2])

    def test_copy_region_errors(self, context: AbstractContext,
                                command_queue: AbstractCommandQueue) -> None:
        # Too many axes
        with pytest.raises(IndexError):
            self._test_copy_region(
                context, command_queue, (10,), None, (10,), None,
                np.s_[3, 4], np.s_[5, 6])
        # Out-of-range single index
        with pytest.raises(IndexError):
            self._test_copy_region(
                context, command_queue, (10,), None, (10,), None,
                np.s_[5], np.s_[10])
        # Out-of-range slice
        with pytest.raises(IndexError):
            self._test_copy_region(
                context, command_queue, (10,), None, (10,), None,
                np.s_[10:12], np.s_[8:10])
        # Empty slice
        with pytest.raises(IndexError):
            self._test_copy_region(
                context, command_queue, (10,), None, (10,), None,
                np.s_[2:2], np.s_[3:3])
        # Negative stride
        with pytest.raises(IndexError):
            self._test_copy_region(
                context, command_queue, (10,), None, (10,), None,
                np.s_[3:0:-1], np.s_[4:1:-1])

    def test_zero(self, array: DeviceArray, context: AbstractContext,
                  command_queue: AbstractCommandQueue) -> None:
        ary = np.random.randint(0x12345678, 0x23456789, array.shape).astype(np.int32)
        array.set(command_queue, ary)
        before = array.get(command_queue)
        array.zero(command_queue)
        after = array.get(command_queue)
        np.testing.assert_equal(ary, before)
        assert np.max(after) == 0

    def _allocate_raw(self, context: AbstractContext, n_bytes: int) -> Any:
        return context.allocate_raw(n_bytes)

    def test_raw(self, shape: Tuple[int, ...], padded_shape: Tuple[int, ...],
                 context: AbstractContext) -> None:
        raw = self._allocate_raw(context, 2048)
        ary = self.cls(
            context=context,
            shape=shape,
            dtype=np.int32,
            padded_shape=padded_shape,
            raw=raw)
        actual_raw = None
        try:
            # CUDA
            if isinstance(ary, SVMArray):
                actual_raw = ary.base.base  # type: ignore
                raw = raw._wrapped
            else:
                actual_raw = ary.buffer.gpudata
        except AttributeError:
            # OpenCL
            actual_raw = ary.buffer.data
        assert actual_raw is raw


class TestPinnedAMD(TestDeviceArray):
    """Run DeviceArray tests forcing `_PinnedAMD` class for pinned memory."""

    @pytest.fixture(autouse=True)
    def setup(self, context: AbstractContext):
        context._force_pinned_amd = True        # type: ignore
        yield
        context._force_pinned_amd = False       # type: ignore


@pytest.mark.cuda_only
class TestSVMArrayHost(TestHostArray):
    """Test SVMArray using the HostArray tests."""

    cls = SVMArray

    # Function is given a different name to avoid mypy complaining about the
    # signature being different.
    @pytest.fixture(name='factory')
    def factory_override(self, context: AbstractContext) -> \
            Callable[[Tuple[int, ...], DTypeLike, Tuple[int, ...]], HostArray]:
        def allocate(shape: Tuple[int, ...], dtype: DTypeLike,
                     padded_shape: Tuple[int, ...]) -> HostArray:
            return self.cls(context, shape, dtype, padded_shape)
        return allocate


@pytest.mark.cuda_only
class TestSVMArray(TestDeviceArray):
    """Tests SVMArray using the DeviceArray tests, plus some new ones."""

    cls = SVMArray

    def _allocate_raw(self, context: AbstractContext, n_bytes: int) -> object:
        return context.allocate_svm_raw(n_bytes)

    def test_coherence(self, context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        """Check that runtime correct transfers data between the host and device views."""
        source = """\
            <%include file="/port.mako"/>
            KERNEL void triple(unsigned int *data)
            {
                data[get_global_id(0)] *= 3;
            }"""
        program = accel.build(context, 'not a file', source=source)
        kernel = program.get_kernel("triple")
        ary = SVMArray(context, (123,), np.uint32, (128,))
        ary[:] = np.arange(123, dtype=np.uint32)
        command_queue.enqueue_kernel(kernel, [ary.buffer], (128,), (64,))
        command_queue.finish()
        np.testing.assert_equal(np.arange(0, 369, step=3, dtype=np.uint32), ary)


@pytest.mark.cuda_only
class TestSVMAllocator:
    def test_allocate(self, context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        allocator = accel.SVMAllocator(context)
        ary = allocator.allocate((12, 34), np.int32, (13, 35))
        assert isinstance(ary, accel.SVMArray)
        assert ary.shape == (12, 34)
        assert ary.dtype == np.int32
        assert ary.padded_shape == (13, 35)
        ary.fill(1)  # Just to make sure it doesn't crash

    def test_allocate_raw(self, context: AbstractContext,
                          command_queue: AbstractCommandQueue) -> None:
        allocator = accel.SVMAllocator(context)
        raw = allocator.allocate_raw(13 * 35 * 4)
        ary = allocator.allocate((12, 34), np.int32, (13, 35), raw=raw)
        assert isinstance(ary, accel.SVMArray)
        assert ary.shape == (12, 34)
        assert ary.dtype == np.int32
        assert ary.padded_shape == (13, 35)
        ary.fill(1)  # Just to make sure it doesn't crash


class TestDimension:
    """Tests for :class:`katsdpsigproc.accel.Dimension`."""

    def test_is_power2(self) -> None:
        assert accel.Dimension._is_power2(1)
        assert accel.Dimension._is_power2(2)
        assert accel.Dimension._is_power2(32)
        assert not accel.Dimension._is_power2(-1)
        assert not accel.Dimension._is_power2(0)
        assert not accel.Dimension._is_power2(3)
        assert not accel.Dimension._is_power2(5)

    def test_min_padded_round(self) -> None:
        """Constructor computes min padded size correctly."""
        dim = accel.Dimension(17, min_padded_round=4)
        assert dim.min_padded_size == 20
        dim = accel.Dimension(20, min_padded_round=5)
        assert dim.min_padded_size == 20

    def test_add_align_dtype(self) -> None:
        dim = accel.Dimension(20, alignment=8)
        assert dim.alignment == 8
        dim.add_align_dtype(np.complex64)
        assert dim.alignment_hint == accel.Dimension.ALIGN_BYTES // 8
        dim.add_align_dtype(np.uint8)
        assert dim.alignment_hint == accel.Dimension.ALIGN_BYTES
        dim.add_align_dtype(np.float32)
        assert dim.alignment_hint == accel.Dimension.ALIGN_BYTES

    def test_valid(self) -> None:
        """Test `valid` method on non-exact dimension."""
        dim = accel.Dimension(17, min_padded_round=8, alignment=4)
        assert dim.valid(24)
        assert not dim.valid(20)
        assert dim.valid(28)
        assert not dim.valid(30)

    def test_valid_exact(self) -> None:
        """Test `valid` method on exact dimension."""
        dim = accel.Dimension(20, alignment=4, exact=True)
        assert dim.valid(20)
        assert not dim.valid(24)
        dim = accel.Dimension(20, min_padded_size=23, exact=True)
        assert dim.valid(23)
        assert not dim.valid(24)
        assert not dim.valid(20)

    @classmethod
    def assert_dimensions_equal(cls, dim1: accel.Dimension, dim2: accel.Dimension) -> None:
        assert dim1.size == dim2.size
        assert dim1.min_padded_size == dim2.min_padded_size
        assert dim1.alignment == dim2.alignment
        assert dim1.alignment_hint == dim2.alignment_hint
        assert dim1.exact == dim2.exact

    def test_link(self) -> None:
        """Linking several dimensions together gives them all the right properties."""
        dim1 = accel.Dimension(22, min_padded_size=28, alignment=4)
        dim2 = accel.Dimension(22, min_padded_size=24, alignment=8, align_dtype=np.int32)
        dim3 = accel.Dimension(22, min_padded_size=22, align_dtype=np.uint16)
        dim1.link(dim2)
        dim1.link(dim3)
        self.assert_dimensions_equal(dim1, dim2)
        self.assert_dimensions_equal(dim2, dim3)
        assert dim1.size == 22
        assert dim1.min_padded_size == 28
        assert dim1.alignment == 8
        assert dim1.alignment_hint == accel.Dimension.ALIGN_BYTES // 2
        assert dim1.exact is False

    def test_link_bad_size(self) -> None:
        """Linking dimensions with different sizes fails."""
        dim1 = accel.Dimension(22, min_padded_size=28, alignment=4)
        dim2 = accel.Dimension(23, min_padded_size=24, alignment=8, align_dtype=np.int32)
        with pytest.raises(ValueError):
            dim1.link(dim2)
        assert dim1._root() is not dim2._root()

    def test_link_bad_exact(self) -> None:
        """Linking dimensions into an unsatisfiable requirement fails."""
        dim1 = accel.Dimension(22, exact=True)
        dim2 = accel.Dimension(22, min_padded_size=28)
        dim3 = accel.Dimension(22, alignment=4)
        with pytest.raises(ValueError):
            dim1.link(dim2)
        with pytest.raises(ValueError):
            dim1.link(dim3)
        # Check that linking didn't happen anyway
        assert dim1._root() is not dim2._root()
        assert dim1._root() is not dim3._root()

    def test_required_padded_size(self) -> None:
        """The padded size is computed correctly."""
        dim = accel.Dimension(30, 7, alignment=4)
        assert dim.required_padded_size() == 36

    def test_required_padded_size_dtype(self) -> None:
        """The padded size is computed correctly when an alignment hint is given."""
        dim = accel.Dimension(1100, 200, align_dtype=np.float32)
        assert dim.required_padded_size() == 1216

    def test_required_padded_size_exact(self) -> None:
        """The padded size is computed correctly for exact dimensions."""
        dim = accel.Dimension(1100, align_dtype=np.float32, exact=True)
        assert dim.required_padded_size() == 1100

    def test_required_padded_size_small(self) -> None:
        """The alignment hint is ignored for small sizes."""
        dim = accel.Dimension(18, alignment=8, align_dtype=np.uint8)
        assert dim.required_padded_size() == 24


class TestIOSlot:
    """Tests for :class:`katsdpsigproc.accel.IOSlot`."""

    @pytest.mark.parametrize("bind", [True, False])
    @mock.patch('katsdpsigproc.accel.DeviceArray', spec=True)
    def test_allocate(self, DeviceArray: mock.Mock, bind: bool) -> None:
        """IOSlot.allocate must correctly allocate a buffer."""
        dims = (
            accel.Dimension(50, min_padded_size=60, alignment=8),
            accel.Dimension(30, min_padded_size=50, alignment=4)
        )
        shape = (50, 30)
        padded_shape = (64, 52)
        dtype = np.dtype(np.uint8)
        # Create the device array that will be created. We need to populate it
        # with some attributes to allow validation to pass
        ary = mock.Mock()
        ary.dtype = dtype
        ary.shape = shape
        ary.padded_shape = padded_shape
        # Set the mocked DeviceArray class to return this array
        DeviceArray.return_value = ary
        # Run the system under test
        slot = accel.IOSlot(dims, dtype)
        ret = slot.allocate(accel.DeviceAllocator(mock.sentinel.context), bind=bind)
        # Validation
        assert ret == ary
        if bind:
            assert slot.buffer == ary
        else:
            assert slot.buffer is None
        DeviceArray.assert_called_once_with(
            mock.sentinel.context, shape, dtype, padded_shape, None)
        # Check that the inner dimension had a type hint set
        assert dims[1].alignment_hint == accel.Dimension.ALIGN_BYTES

    @mock.patch('katsdpsigproc.accel.HostArray', spec=True)
    def test_allocate_host(self, HostArray: mock.Mock) -> None:
        dims = (
            accel.Dimension(50, min_padded_size=60, alignment=8),
            accel.Dimension(30, min_padded_size=50, alignment=4)
        )
        shape = (50, 30)
        padded_shape = (64, 52)
        dtype = np.int16
        slot = accel.IOSlot(dims, dtype)
        HostArray.return_value = mock.sentinel.host_array
        ret = slot.allocate_host(mock.sentinel.context)
        assert ret is mock.sentinel.host_array
        HostArray.assert_called_with(shape, dtype, padded_shape, context=mock.sentinel.context)

    @mock.patch('katsdpsigproc.accel.DeviceArray', spec=True)
    def test_allocate_raw(self, DeviceArray: mock.Mock) -> None:
        """Test IOSlot.allocate with a raw parameter."""
        shape = (50, 30)
        dtype = np.dtype(np.float32)
        raw = mock.sentinel.raw
        # Create the device array that will be created. We need to populate it
        # with some attributes to allow validation to pass
        ary = mock.Mock()
        ary.dtype = dtype
        ary.shape = shape
        ary.padded_shape = shape
        ary.raw = raw
        # Set the mocked DeviceArray class to return this array
        DeviceArray.return_value = ary
        # Run the system under test
        slot = accel.IOSlot(shape, dtype)
        slot.allocate(accel.DeviceAllocator(mock.sentinel.context), raw)
        # Validation
        assert slot.buffer == ary
        DeviceArray.assert_called_once_with(
            mock.sentinel.context, shape, dtype, shape, raw)

    def test_validate_shape(self) -> None:
        """IOSlot.validate must check that the shape matches."""
        ary = mock.sentinel.ary
        ary.dtype = np.float32
        ary.shape = (5, 3)
        ary.padded_shape = (5, 3)
        slot = accel.IOSlot((5, 3), np.float32)
        slot.bind(ary)  # Should pass

        with pytest.raises(ValueError):
            # Wrong shape
            ary.shape = (5, 4)
            slot.bind(ary)
        with pytest.raises(ValueError):
            # Wrong dimensions
            ary.shape = (5,)
            slot.bind(ary)

    def test_validate_dtype(self) -> None:
        """IOSlot.validate must check that the dtype matches."""
        ary = mock.sentinel.ary
        ary.dtype = np.float32
        ary.shape = (5, 3)
        ary.padded_shape = ary.shape
        slot = accel.IOSlot((5, 3), np.float32)
        slot.bind(ary)  # Should pass

        ary.dtype = np.dtype(np.float32)  # Equivalent dtype
        slot.bind(ary)  # Should pass

        with pytest.raises(TypeError):
            ary.dtype = np.int32
            slot.bind(ary)

    def test_validate_padded_shape(self) -> None:
        """IOSlot.validate must check that the padded shape is valid."""
        ary = mock.sentinel.ary
        ary.dtype = np.float32
        ary.shape = (5, 3)
        ary.padded_shape = (8, 6)
        slot = accel.IOSlot((accel.Dimension(5, 8), accel.Dimension(3, 6)), np.float32)
        slot.bind(ary)  # Should pass

        ary.padded_shape = (10, 11)
        with pytest.raises(ValueError):
            # Bigger than needed, but not an exact match - must fail
            slot.bind(ary)

    def test_required_bytes(self) -> None:
        slot = accel.IOSlot(
            (accel.Dimension(27, alignment=4), accel.Dimension(33, alignment=32)),
            np.float32)
        assert slot.required_bytes() == 4 * 28 * 64

    def test_bind_none(self) -> None:
        """IOSlot.bind must accept `None`."""
        ary = mock.sentinel.ary
        ary.dtype = np.float32
        ary.shape = (5, 3)
        ary.padded_shape = ary.shape
        slot = accel.IOSlot((5, 3), np.float32)
        slot.bind(ary)
        assert slot.buffer is ary
        slot.bind(None)
        assert slot.buffer is None


class TestCompoundIOSlot:
    """Tests for :class:`katsdpsigproc.accel.CompoundIOSlot`."""

    def setup_method(self) -> None:
        self.dims1 = (
            accel.Dimension(13, min_padded_size=17, alignment=1),
            accel.Dimension(7, min_padded_size=8, alignment=8),
            accel.Dimension(22, min_padded_size=25, alignment=4)
        )
        self.dims2 = (
            accel.Dimension(13, min_padded_size=14, alignment=4),
            accel.Dimension(7, min_padded_size=10, alignment=4),
            accel.Dimension(22, min_padded_size=22, alignment=1)
        )
        self.slot1 = accel.IOSlot(self.dims1, np.float32)
        self.slot2 = accel.IOSlot(self.dims2, np.float32)

    def test_check_empty(self) -> None:
        """:class:`.CompoundIOSlot` constructor must reject empty list."""
        with pytest.raises(ValueError):
            accel.CompoundIOSlot([])

    def test_validate_shape(self) -> None:
        """:class:`.CompoundIOSlot` must check that children have consistent shapes."""
        slot1 = accel.IOSlot((5, 3), np.float32)
        slot2 = accel.IOSlot((5, 4), np.float32)
        with pytest.raises(ValueError):
            accel.CompoundIOSlot([slot1, slot2])

    def test_validate_dtype(self) -> None:
        """:class:`.CompoundIOSlot` must check that children have consistent data types."""
        slot1 = accel.IOSlot((5, 3), np.float32)
        slot2 = accel.IOSlot((5, 3), np.int32)
        with pytest.raises(TypeError):
            accel.CompoundIOSlot([slot1, slot2])

    def test_attributes(self) -> None:
        """:class:`.CompoundIOSlot` must correctly combine attributes."""
        slot = accel.CompoundIOSlot([self.slot1, self.slot2])
        assert slot.shape == (13, 7, 22)
        assert slot.dtype == np.float32
        assert self.dims1[0].min_padded_size == 17
        assert self.dims1[1].min_padded_size == 10
        assert self.dims1[2].min_padded_size == 25
        assert self.dims1[0].alignment == 4
        assert self.dims1[1].alignment == 8
        assert self.dims1[2].alignment == 4
        for x, y in zip(self.dims1, self.dims2):
            TestDimension.assert_dimensions_equal(x, y)

    def test_bind(self) -> None:
        """CompoundIOSlot.bind must bind children."""
        ary = mock.sentinel.ary
        ary.shape = (13, 7, 22)
        ary.dtype = np.float32
        ary.padded_shape = (20, 16, 28)
        slot = accel.CompoundIOSlot([self.slot1, self.slot2])
        slot.bind(ary)
        assert slot.buffer == ary
        assert self.slot1.buffer == ary
        assert self.slot2.buffer == ary

    def test_bind_fail(self) -> None:
        """CompoundIOSlot.bind must not have side effects if the array is \
        not valid for all children."""
        ary = mock.sentinel.ary
        ary.shape = (13, 7, 22)
        ary.dtype = np.float32
        ary.padded_shape = (17, 8, 28)
        # Check that the array has the properties we want for the test
        self.slot1.validate(ary)
        with pytest.raises(ValueError):
            self.slot2.validate(ary)
        # The actual test
        slot = accel.CompoundIOSlot([self.slot1, self.slot2])
        with pytest.raises(ValueError):
            slot.bind(ary)
        assert slot.buffer is None
        assert self.slot1.buffer is None
        assert self.slot2.buffer is None


class TestAliasIOSlot:
    """Tests for :class:`katsdpsigproc.accel.AliasIOSlot`."""

    def setup_method(self) -> None:
        self.slot1 = accel.IOSlot((3, 7), np.float32)
        self.slot2 = accel.IOSlot((5, 3), np.complex64)

    def test_required_bytes(self) -> None:
        slot = accel.AliasIOSlot([self.slot1, self.slot2])
        assert slot.required_bytes() == 120

    def test_allocate(self) -> None:
        # Set up mocks
        raw = mock.sentinel.raw
        context = mock.NonCallableMock()
        context.allocate_raw = mock.Mock(return_value=raw)
        # Run the test
        slot = accel.AliasIOSlot([self.slot1, self.slot2])
        ret = slot.allocate(accel.DeviceAllocator(context))
        # Validation
        context.allocate_raw.assert_called_once_with(120)
        assert ret is raw
        assert slot.raw is raw
        assert self.slot1.buffer is not None
        assert self.slot2.buffer is not None

    @mock.patch('katsdpsigproc.accel.HostArray', spec=True)
    def test_allocate_host(self, HostArray: mock.Mock) -> None:
        HostArray.return_value = mock.sentinel.host_array
        slot = accel.AliasIOSlot([self.slot1, self.slot2])
        ret = slot.allocate_host(mock.sentinel.context)
        assert ret is mock.sentinel.host_array
        HostArray.assert_called_with((120,), np.uint8, context=mock.sentinel.context)


class TestVisualizeOperation:
    """Tests for :class:`katsdpsigproc.accel.visualize_operation`.

    This is just a basic smoke test to ensure that it runs without crashing.
    A real test requires a human to sanity-check the visual output.
    """

    class Inner(accel.Operation):
        def _run(self) -> None:
            pass

    def setup_method(self) -> None:
        self.tmpdir = tempfile.mkdtemp()

    def teardown_method(self) -> None:
        shutil.rmtree(self.tmpdir)

    def test(self, context: AbstractContext, command_queue: AbstractCommandQueue) -> None:
        inner = TestVisualizeOperation.Inner(command_queue)
        inner.slots['foo'] = accel.IOSlot((3, 4), np.float32)
        inner.slots['bar'] = accel.IOSlot((4, 3), np.float32)

        mid = accel.OperationSequence(
            command_queue,
            [('inner', inner)],
            {'inner_foo': ['inner:foo']})

        outer = accel.OperationSequence(
            command_queue,
            [('mid', mid)],
            {'mid_foo': ['mid:inner_foo']},
            {'scratch': ['mid:inner:bar']})

        accel.visualize_operation(outer, os.path.join(self.tmpdir, 'test_visualize.gv'))



################################################################################
# Copyright (c) 2014-2022, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for rank.mako."""

import os
from typing import Any

import numpy as np
try:
    from numpy.typing import DTypeLike, ArrayLike
except ImportError:
    DTypeLike = Any     # type: ignore
    ArrayLike = Any   # type: ignore
import pytest

from katsdpsigproc import accel
from katsdpsigproc.accel import DeviceArray, build
from katsdpsigproc.abc import AbstractContext, AbstractCommandQueue, AbstractProgram


_wgs = 128
_M = 1000
_N = 2000


@pytest.fixture(scope="module")
def data() -> np.ndarray:
    rs = np.random.RandomState(seed=1)   # Fixed seed to make test repeatable
    return rs.randint(0, _M, size=_N).astype(np.int32)


@pytest.fixture(scope="module")
def expected(data: np.ndarray) -> np.ndarray:
    expected = np.empty(_M, dtype=np.int32)
    for i in range(_M):
        expected[i] = np.sum(np.less(data, i))
    return expected


@pytest.fixture
def program(context: AbstractContext) -> AbstractProgram:
    return build(
        context, 'test_rank.mako', {'size': _wgs},
        extra_dirs=[os.path.dirname(__file__)])


@pytest.mark.parametrize(
    'kernel_name, wgs',
    [
        ('test_rank_serial', 1),
        ('test_rank_parallel', _wgs)
    ]
)
def test_rank(context: AbstractContext, command_queue: AbstractCommandQueue,
              data: np.ndarray, expected: np.ndarray, program: AbstractProgram,
              kernel_name: str, wgs: int) -> None:
    data_d = DeviceArray(context, shape=data.shape, dtype=data.dtype)
    data_d.set(command_queue, data)
    out_d = DeviceArray(context, shape=expected.shape, dtype=expected.dtype)
    kernel = program.get_kernel(kernel_name)
    command_queue.enqueue_kernel(
        kernel, [
            data_d.buffer,
            out_d.buffer,
            np.int32(_N),
            np.int32(_M)
        ],
        global_size=(wgs,), local_size=(wgs,))
    out = out_d.get(command_queue)
    np.testing.assert_equal(expected, out)


def run_float_func(context: AbstractContext, command_queue: AbstractCommandQueue,
                   program: AbstractProgram,
                   kernel_name: str, data: np.ndarray, output_size: int) -> np.ndarray:
    """Do common testing for find_min_float, find_max_float, median_non_zero_float."""
    size = 128  # Number of workitems
    kernel = program.get_kernel(kernel_name)
    data = data.astype(np.float32)
    N = data.shape[0]
    data_d = accel.DeviceArray(context, shape=data.shape, dtype=np.float32)
    data_d.set(command_queue, data)
    out_d = accel.DeviceArray(context, shape=(output_size,), dtype=np.float32)
    command_queue.enqueue_kernel(
        kernel,
        [data_d.buffer, out_d.buffer, np.int32(N)],
        global_size=(size,),
        local_size=(size,))
    out = out_d.empty_like()
    out_d.get(command_queue, out)
    return out


class TestFindMinMaxFloat:
    def check_array(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                    program: AbstractProgram, data_: ArrayLike) -> None:
        data = np.asarray(data_, dtype=np.float32)
        expected = [np.nanmin(data), np.nanmax(data)]
        out = run_float_func(context, command_queue, program, 'test_find_min_max_float', data, 2)
        assert expected[0] == out[0]
        assert expected[1] == out[1]

    @pytest.mark.parametrize('data', [[5.3], [-10.0, 5.5, np.nan, -20.0, np.nan]])
    def test_simple(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                    program: AbstractProgram, data: ArrayLike) -> None:
        self.check_array(context, command_queue, program, data)

    @pytest.mark.parametrize('ordered', [False, True])
    def test_random(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                    program: AbstractProgram, ordered: bool) -> None:
        rs = np.random.RandomState(seed=1)   # Fixed seed to make test repeatable
        data = rs.uniform(-10.0, 10.0, 1000)
        if ordered:
            data = np.sort(data)
        self.check_array(context, command_queue, program, data)

    def test_all_nan(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                     program: AbstractProgram) -> None:
        # Can't use check_array, because np.nanmin warns if all are NaN,
        # and asserting == doesn't handle NaN.
        data = np.array([np.nan, np.nan], dtype=np.float32)
        out = run_float_func(context, command_queue, program, 'test_find_min_max_float', data, 2)
        assert np.isnan(out[0])
        assert np.isnan(out[1])


class TestMedianNonZero:
    def check_array(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                    program: AbstractProgram, data_: ArrayLike) -> None:
        data = np.asarray(data_, dtype=np.float32)
        expected = np.median(data[data > 0.0])
        out = run_float_func(context, command_queue, program, 'test_median_non_zero', data, 2)
        assert expected == out[0]
        assert expected == out[1]

    @pytest.mark.parametrize(
        'data',
        [
            [5.3],
            [1.2, 2.3, 3.4, 4.5],
            [0.0, 0.0, 0.0, 1.2, 5.3, 2.4]
        ]
    )
    def test_simple(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                    program: AbstractProgram, data: ArrayLike) -> None:
        self.check_array(context, command_queue, program, data)

    def test_big_even(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                      program: AbstractProgram) -> None:
        rs = np.random.RandomState(seed=1)   # Fixed seed to make test repeatable
        data = rs.random_sample(10001) + 0.5
        data[123] = 0.0
        self.check_array(context, command_queue, program, data)

    def test_big_odd(self, context: AbstractContext, command_queue: AbstractCommandQueue,
                     program: AbstractProgram) -> None:
        rs = np.random.RandomState(seed=2)   # Fixed seed to make test repeatable
        data = rs.random_sample(10000) + 0.5
        data[456] = 0.0
        self.check_array(context, command_queue, program, data)



#!/usr/bin/env python

################################################################################
# Copyright (c) 2014-2020, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Test script that runs RFI flagging on random or real data."""

import argparse
import time
import sys
import json
import concurrent.futures

import numpy as np

import katsdpsigproc.rfi.host
import katsdpsigproc.rfi.device
import katsdpsigproc.rfi.twodflag
import katsdpsigproc.accel as accel


def generate_data(times, channels, baselines):
    rs = np.random.RandomState(seed=1)
    shape = (channels, baselines) if times is None else (times, channels, baselines)
    # This is done a row at a time to keep memory usage low
    out = np.empty(shape, np.complex64)
    for i in range(shape[0]):
        real = rs.standard_normal(size=shape[1:]).astype(np.float32)
        imag = rs.standard_normal(size=shape[1:]).astype(np.float32)
        out[i] = real + 1j * imag
    return out


def benchmark1d(args, data):
    if args.width % 2 != 1:
        raise argparse.ArgumentError('Width must be odd')
    if data.shape[0] <= args.width:
        raise argparse.ArgumentError('Channels cannot be less than the filter width')

    context = None
    if not args.host:
        try:
            context = accel.create_some_context(True)
        except RuntimeError:
            print("No devices available. Executing on the CPU.", file=sys.stderr)

    if context is None:
        background = katsdpsigproc.rfi.host.BackgroundMedianFilterHost(args.width)
        noise_est = katsdpsigproc.rfi.host.NoiseEstMADHost()
        threshold = katsdpsigproc.rfi.host.ThresholdSumHost(args.sigmas)
        flagger = katsdpsigproc.rfi.host.FlaggerHost(background, noise_est, threshold)
        start = time.time()
        flags = flagger(data)
        end = time.time()
        print("CPU time (ms):", (end - start) * 1000.0)
    else:
        command_queue = context.create_command_queue(profile=True)
        background = katsdpsigproc.rfi.device.BackgroundMedianFilterDeviceTemplate(
            context, args.width)
        noise_est = katsdpsigproc.rfi.device.NoiseEstMADTDeviceTemplate(
            context, 10240)
        threshold = katsdpsigproc.rfi.device.ThresholdSumDeviceTemplate(context)
        template = katsdpsigproc.rfi.device.FlaggerDeviceTemplate(background, noise_est, threshold)
        flagger = template.instantiate(command_queue, data.shape[0], data.shape[1],
                                       threshold_args={'n_sigma': args.sigmas})
        flagger.ensure_all_bound()

        data_device = flagger.buffer('vis')
        flags_device = flagger.buffer('flags')

        data_device.set(command_queue, data)
        # Run once for warmup (allocates memory)
        flagger()
        # Run again, timing it
        command_queue.finish()

        start_time = time.time()
        start_event = command_queue.enqueue_marker()
        flagger()
        end_event = command_queue.enqueue_marker()
        command_queue.finish()
        end_time = time.time()
        flags = flags_device.get(command_queue)
        print("Host time (ms):  ", (end_time - start_time) * 1000.0)
        try:
            device_time = end_event.time_since(start_event) * 1000.0
        except Exception:
            # AMD CPU device doesn't seem to support profiling on marker events
            device_time = 'unknown'
        print("Device time (ms):", device_time)
    return flags


def benchmark2d(args, data):
    flagger = katsdpsigproc.rfi.twodflag.SumThresholdFlagger(**args.params)
    in_flags = np.zeros(data.shape, np.bool_)
    if args.pool == 'none':
        pool = None
    elif args.pool == 'process':
        pool = concurrent.futures.ProcessPoolExecutor(args.workers)
    elif args.pool == 'thread':
        pool = concurrent.futures.ThreadPoolExecutor(args.workers)
    else:
        raise argparse.ArgumentError(f'unhandled value {args.pool} for --pool')
    # Warmup
    try:
        flagger.get_flags(data[:2], in_flags[:2], pool=pool)
        start = time.time()
        flags = flagger.get_flags(data, in_flags, pool=pool)
        end = time.time()
        print("CPU time (ms):", (end - start) * 1000.0)
    finally:
        if pool is not None:
            pool.shutdown()
    return flags


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('--host', action='store_true',
                        help='Compute on the host instead of with CUDA/OpenCL')

    parser.add_argument_group('Data selection')
    parser.add_argument('--antennas', '-a', type=int, default=7)
    parser.add_argument('--channels', '-c', type=int, default=1024)
    parser.add_argument('--baselines', '-b', type=int, help='(overrides --antennas)')
    parser.add_argument('--times', '-t', type=int, help='time samples (uses 2D flagger)')
    parser.add_argument('--preset', '-p', choices=('small', 'medium', 'big', 'kat7'),
                        help='(overrides other options)')
    parser.add_argument('--file', type=str, help='specify a real data file (.npy)')

    parser.add_argument_group('Parameters (1D flagger)')
    parser.add_argument('--width', '-w', type=int, help='median filter kernel size (must be odd)',
                        default=13)
    parser.add_argument('--sigmas', type=float, help='threshold for detecting RFI', default=11.0)

    parser.add_argument_group('Parameters (2D flagger)')
    parser.add_argument('--params', type=json.loads, default={}, help='JSON dict of parameters')
    parser.add_argument('--pool', choices=('none', 'process', 'thread'),
                        help='parallelization method')
    parser.add_argument('--workers', type=int, help='Number of parallel workers')

    args = parser.parse_args()

    if args.file is not None:
        if args.file.endswith('.npy'):
            data = np.load(args.file)
        else:
            raise argparse.ArgumentError("Don't know how to handle " + args.file)
    else:
        if args.preset is not None:
            args.baselines = None    # Force recomputation from antennas
            if args.preset == 'small':
                args.antennas = 2
                args.channels = max(2 * args.width, 8)
            elif args.preset == 'medium':
                args.antennas = 15
                args.channels = max(2 * args.width, 128)
            elif args.preset == 'kat7':
                args.antennas = 7
                args.channels = 8192
            elif args.preset == 'big':
                args.antennas = 64
                args.channels = 10240
            else:
                raise argparse.ArgumentError('Unexpected value for preset')
        if args.baselines is None:
            # Note: * 2 not / 2 because there are 4 polarisations
            args.baselines = args.antennas * (args.antennas + 1) * 2
        data = generate_data(args.times, args.channels, args.baselines)

    if args.times is None:
        flags = benchmark1d(args, data)
    else:
        flags = benchmark2d(args, data)
    print('{:.4f}% flagged'.format(100.0 * np.sum(flags) / flags.size))


if __name__ == '__main__':
    main()



################################################################################
# Copyright (c) 2021, National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Unit tests for :mod:`katsdpsigproc`."""

from typing import Any, Tuple, Union

import numpy as np
try:
    from numpy.typing import ArrayLike
except ImportError:
    ArrayLike = Any  # type: ignore


_RS = Union[np.random.RandomState, np.random.Generator]


def complex_normal(
    state: _RS,
    loc: ArrayLike = 0.0j,
    scale: ArrayLike = 1.0,
    size: Union[int, Tuple[int, ...], None] = None
):
    """Generate a circularly symmetric Gaussian in the Argand plane."""
    return (
        state.normal(np.real(loc), scale, size)         # type: ignore
        + 1j * state.normal(np.imag(loc), scale, size)  # type: ignore
    )
